# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import re
import keyword
try:
    from collections.abc import Sequence
except:
    from collections import Sequence

import numpy

import awkward1._numpy
import awkward1._pandas
import awkward1.layout
import awkward1.operations.convert

_dir_pattern = re.compile(r"^[a-zA-Z_]\w*$")

class Array(awkward1._numpy.NDArrayOperatorsMixin, awkward1._pandas.PandasMixin, Sequence):
    def __init__(self, data, behavior=None):
        if isinstance(data, awkward1.layout.Content):
            layout = data
        elif isinstance(data, Array):
            layout = data.layout
        elif isinstance(data, numpy.ndarray):
            layout = awkward1.operations.convert.fromnumpy(data).layout
        elif isinstance(data, str):
            layout = awkward1.operations.convert.fromjson(data).layout
        else:
            layout = awkward1.operations.convert.fromiter(data).layout
        if not isinstance(layout, awkward1.layout.Content):
            raise TypeError("could not convert data into an awkward1.Array")

        if self.__class__ is Array:
            self.__class__ = awkward1._util.arrayclass(layout, behavior)

        self.layout = layout
        self.behavior = behavior

    @property
    def layout(self):
        return self._layout

    @layout.setter
    def layout(self, layout):
        if isinstance(layout, awkward1.layout.Content):
            self._layout = layout
            self._numbaview = None
        else:
            raise TypeError("layout must be a subclass of awkward1.layout.Content")

    @property
    def behavior(self):
        return self._behavior

    @behavior.setter
    def behavior(self, behavior):
        if behavior is None or isinstance(behavior, dict):
            self._behavior = behavior
        else:
            raise TypeError("behavior must be None or a dict")

    @property
    def type(self):
        return awkward1.types.ArrayType(self._layout.type, len(self._layout))

    def __len__(self):
        return len(self._layout)

    def __iter__(self):
        for x in self._layout:
            yield awkward1._util.wrap(x, self._behavior)

    def __getitem__(self, where):
        return awkward1._util.wrap(self._layout[where], self._behavior)

    def __setitem__(self, where, what):
        if not isinstance(where, str):
            raise ValueError("only fields may be assigned in-place (by field name)")
        self._layout = awkward1.operations.structure.withfield(self._layout, what, where).layout
        self._numbaview = None

    def __getattr__(self, where):
        if where in dir(super(Array, self)):
            return super(Array, self).__getattribute__(where)
        else:
            if where in self._layout.keys():
                try:
                    return self[where]
                except Exception as err:
                    raise AttributeError("while trying to get field {0}, an exception occurred:\n{1}: {2}".format(repr(where), type(err), str(err)))
            else:
                raise AttributeError("no field named {0}".format(repr(where)))

    def __dir__(self):
        return sorted(set(dir(super(Array, self)) + [x for x in self._layout.keys() if _dir_pattern.match(x) and not keyword.iskeyword(x)]))

    @property
    def i0(self):
        return self["0"]
    @property
    def i1(self):
        return self["1"]
    @property
    def i2(self):
        return self["2"]
    @property
    def i3(self):
        return self["3"]
    @property
    def i4(self):
        return self["4"]
    @property
    def i5(self):
        return self["5"]
    @property
    def i6(self):
        return self["6"]
    @property
    def i7(self):
        return self["7"]
    @property
    def i8(self):
        return self["8"]
    @property
    def i9(self):
        return self["9"]

    def __str__(self, limit_value=85):
        return awkward1._util.minimally_touching_string(limit_value, self._layout, self._behavior)

    def __repr__(self, limit_value=40, limit_total=85):
        value = awkward1._util.minimally_touching_string(limit_value, self._layout, self._behavior)

        limit_type = limit_total - len(value) - len("<Array  type=>")
        type = repr(str(self.type))
        if len(type) > limit_type:
            type = type[:(limit_type - 4)] + "..." + type[-1]

        return "<Array {0} type={1}>".format(value, type)

    def __array__(self, *args, **kwargs):
        if awkward1._util.called_by_module("pandas"):
            try:
                return awkward1._numpy.convert_to_array(self._layout, args, kwargs)
            except:
                out = numpy.empty(len(self._layout), dtype="O")
                for i, x in enumerate(self._layout):
                    out[i] = awkward1._util.wrap(x, self._behavior)
                return out
        else:
            return awkward1._numpy.convert_to_array(self._layout, args, kwargs)

    def __array_function__(self, func, types, args, kwargs):
        return awkward1._numpy.array_function(func, types, args, kwargs)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        return awkward1._numpy.array_ufunc(ufunc, method, inputs, kwargs, self._behavior)

    @property
    def numbatype(self):
        import numba
        import awkward1._numba
        awkward1._numba.register()
        if self._numbaview is None:
            self._numbaview = awkward1._numba.arrayview.ArrayView.fromarray(self)
        return numba.typeof(self._numbaview)

class Record(awkward1._numpy.NDArrayOperatorsMixin):
    def __init__(self, data, behavior=None):
        # FIXME: more checks here
        layout = data
        if not isinstance(layout, awkward1.layout.Record):
            raise TypeError("could not convert data into an awkward1.Record")

        if self.__class__ is Record:
            self.__class__ = awkward1._util.recordclass(layout, behavior)

        self.layout = layout
        self.behavior = behavior

    @property
    def layout(self):
        return self._layout

    @layout.setter
    def layout(self, layout):
        if isinstance(layout, awkward1.layout.Record):
            self._layout = layout
            self._numbaview = None
        else:
            raise TypeError("layout must be a subclass of awkward1.layout.Record")

    @property
    def behavior(self):
        return self._behavior

    @behavior.setter
    def behavior(self, behavior):
        if behavior is None or isinstance(behavior, dict):
            self._behavior = behavior
        else:
            raise TypeError("behavior must be None or a dict")

    @property
    def type(self):
        return self._layout.type

    def __getitem__(self, where):
        return awkward1._util.wrap(self._layout[where], self._behavior)

    def __setitem__(self, where, what):
        if not isinstance(where, str):
            raise ValueError("only fields may be assigned in-place (by field name)")
        self._layout = awkward1.operations.structure.withfield(self._layout, what, where).layout
        self._numbaview = None

    def __getattr__(self, where):
        if where in dir(super(Record, self)):
            return super(Record, self).__getattribute__(where)
        else:
            if where in self._layout.keys():
                try:
                    return self[where]
                except Exception as err:
                    raise AttributeError("while trying to get field {0}, an exception occurred:\n{1}: {2}".format(repr(where), type(err), str(err)))
            else:
                raise AttributeError("no field named {0}".format(repr(where)))

    def __dir__(self):
        return sorted(set(dir(super(Array, self)) + [x for x in self._layout.keys() if _dir_pattern.match(x) and not keyword.iskeyword(x)]))

    @property
    def i0(self):
        return self["0"]
    @property
    def i1(self):
        return self["1"]
    @property
    def i2(self):
        return self["2"]
    @property
    def i3(self):
        return self["3"]
    @property
    def i4(self):
        return self["4"]
    @property
    def i5(self):
        return self["5"]
    @property
    def i6(self):
        return self["6"]
    @property
    def i7(self):
        return self["7"]
    @property
    def i8(self):
        return self["8"]
    @property
    def i9(self):
        return self["9"]

    def __str__(self, limit_value=85):
        return awkward1._util.minimally_touching_string(limit_value + 2, self._layout, self._behavior)[1:-1]

    def __repr__(self, limit_value=40, limit_total=85):
        value = awkward1._util.minimally_touching_string(limit_value + 2, self._layout, self._behavior)[1:-1]

        limit_type = limit_total - len(value) - len("<Record  type=>")
        type = repr(str(self._layout.type))
        if len(type) > limit_type:
            type = type[:(limit_type - 4)] + "..." + type[-1]

        return "<Record {0} type={1}>".format(value, type)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        return awkward1._numpy.array_ufunc(ufunc, method, inputs, kwargs, self._behavior)

    @property
    def numbatype(self):
        import numba
        import awkward1._numba
        awkward1._numba.register()
        if self._numbaview is None:
            self._numbaview = awkward1._numba.arrayview.RecordView.fromrecord(self)
        return numba.typeof(self._numbaview)

class FillableArray(Sequence):
    @classmethod
    def _wrap(cls, fillablearray, behavior=None):
        assert isinstance(fillablearray, awkward1.layout.FillableArray)
        out = cls.__new__(cls)
        out._fillablearray = fillablearray
        out.behavior = behavior
        return out

    def __init__(self, behavior=None):
        self._fillablearray = awkward1.layout.FillableArray()
        self.behavior = behavior

    @property
    def behavior(self):
        return self._behavior

    @behavior.setter
    def behavior(self, behavior):
        if behavior is None or isinstance(behavior, dict):
            self._behavior = behavior
        else:
            raise TypeError("behavior must be None or a dict")

    @property
    def type(self):
        tmp = self._fillablearray.snapshot()
        return awkward1.types.ArrayType(tmp.type, len(tmp))

    def __len__(self):
        return len(self._fillablearray)

    def __getitem__(self, where):
        return awkward1._util.wrap(self._fillablearray[where], self._behavior)

    def __iter__(self):
        for x in self._fillablearray.snapshot():
            yield awkward1._util.wrap(x, self._behavior)

    def __str__(self, limit_value=85, snapshot=None):
        if snapshot is None:
            snapshot = self.snapshot()
        return snapshot.__str__(limit_value=limit_value)

    def __repr__(self, limit_value=40, limit_total=85):
        snapshot = self.snapshot()
        value = self.__str__(limit_value=limit_value, snapshot=snapshot)

        limit_type = limit_total - len(value) - len("<FillableArray  type=>")
        type = repr(str(snapshot.type))
        if len(type) > limit_type:
            type = type[:(limit_type - 4)] + "..." + type[-1]

        return "<FillableArray {0} type={1}>".format(value, type)

    def __array__(self, *args, **kwargs):
        return awkward1._numpy.convert_to_array(self._fillablearray.snapshot(), args, kwargs)

    def __array_function__(self, func, types, args, kwargs):
        return awkward1._numpy.array_function(func, types, args, kwargs)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        return awkward1._numpy.array_ufunc(ufunc, method, inputs, kwargs, self._behavior)

    @property
    def numbatype(self):
        import numba
        import awkward1._numba.fillable
        awkward1._numba.register()
        return awkward1._numba.fillable.FillableArrayType(self._behavior)

    def snapshot(self):
        return awkward1._util.wrap(self._fillablearray.snapshot(), self._behavior)

    def null(self):
        self._fillablearray.null()

    def boolean(self, x):
        self._fillablearray.boolean(x)

    def integer(self, x):
        self._fillablearray.integer(x)

    def real(self, x):
        self._fillablearray.real(x)

    def bytestring(self, x):
        self._fillablearray.bytestring(x)

    def string(self, x):
        self._fillablearray.string(x)

    def beginlist(self):
        self._fillablearray.beginlist()

    def endlist(self):
        self._fillablearray.endlist()

    def begintuple(self):
        self._fillablearray.begintuple()

    def index(self, i):
        self._fillablearray.index(i)

    def endtuple(self):
        self._fillablearray.endtuple()

    def beginrecord(self, name=None):
        self._fillablearray.beginrecord(name)

    def field(self, key):
        self._fillablearray.field(key)

    def endrecord(self):
        self._fillablearray.endrecord()

    def append(self, obj, at=None):
        if at is None:
            if isinstance(obj, Record):
                self._fillablearray.append(obj.layout.array, obj.layout.at)
            elif isinstance(obj, Array):
                self._fillablearray.extend(obj.layout)
            else:
                self._fillablearray.fromiter(obj)

        else:
            if isinstance(obj, Array):
                self._fillablearray.append(obj.layout, at)
            else:
                raise TypeError("'append' method can only be used with 'at' when 'obj' is an ak.Array")

    def extend(self, obj):
        if isinstance(obj, Array):
            self._fillablearray.extend(obj.layout)
        else:
            raise TypeError("'extend' method requires an ak.Array")



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import numpy

import awkward1._util
import awkward1._numpy
import awkward1.layout
import awkward1.operations.convert

def count(array, axis=None, keepdims=False):
    layout = awkward1.operations.convert.tolayout(array, allowrecord=False, allowother=False)
    if axis is None:
        def reduce(xs):
            if len(xs) == 1:
                return xs[0]
            else:
                return xs[0] + reduce(xs[1:])
        return reduce([numpy.size(x) for x in awkward1._util.completely_flatten(layout)])
    else:
        behavior = awkward1._util.behaviorof(array)
        return awkward1._util.wrap(layout.count(axis=axis, mask=False, keepdims=keepdims), behavior)

@awkward1._numpy.implements(numpy.count_nonzero)
def count_nonzero(array, axis=None, keepdims=False):
    layout = awkward1.operations.convert.tolayout(array, allowrecord=False, allowother=False)
    if axis is None:
        def reduce(xs):
            if len(xs) == 1:
                return xs[0]
            else:
                return xs[0] + reduce(xs[1:])
        return reduce([numpy.count_nonzero(x) for x in awkward1._util.completely_flatten(layout)])
    else:
        behavior = awkward1._util.behaviorof(array)
        return awkward1._util.wrap(layout.count_nonzero(axis=axis, mask=False, keepdims=keepdims), behavior)

@awkward1._numpy.implements(numpy.sum)
def sum(array, axis=None, keepdims=False):
    layout = awkward1.operations.convert.tolayout(array, allowrecord=False, allowother=False)
    if axis is None:
        def reduce(xs):
            if len(xs) == 1:
                return xs[0]
            else:
                return xs[0] + reduce(xs[1:])
        return reduce([numpy.sum(x) for x in awkward1._util.completely_flatten(layout)])
    else:
        behavior = awkward1._util.behaviorof(array)
        return awkward1._util.wrap(layout.sum(axis=axis, mask=False, keepdims=keepdims), behavior)

@awkward1._numpy.implements(numpy.prod)
def prod(array, axis=None, keepdims=False):
    layout = awkward1.operations.convert.tolayout(array, allowrecord=False, allowother=False)
    if axis is None:
        def reduce(xs):
            if len(xs) == 1:
                return xs[0]
            else:
                return xs[0] * reduce(xs[1:])
        return reduce([numpy.prod(x) for x in awkward1._util.completely_flatten(layout)])
    else:
        behavior = awkward1._util.behaviorof(array)
        return awkward1._util.wrap(layout.prod(axis=axis, mask=False, keepdims=keepdims), behavior)

@awkward1._numpy.implements(numpy.any)
def any(array, axis=None, keepdims=False):
    layout = awkward1.operations.convert.tolayout(array, allowrecord=False, allowother=False)
    if axis is None:
        def reduce(xs):
            if len(xs) == 1:
                return xs[0]
            else:
                return xs[0] or reduce(xs[1:])
        return reduce([numpy.any(x) for x in awkward1._util.completely_flatten(layout)])
    else:
        behavior = awkward1._util.behaviorof(array)
        return awkward1._util.wrap(layout.any(axis=axis, mask=False, keepdims=keepdims), behavior)

@awkward1._numpy.implements(numpy.all)
def all(array, axis=None, keepdims=False):
    layout = awkward1.operations.convert.tolayout(array, allowrecord=False, allowother=False)
    if axis is None:
        def reduce(xs):
            if len(xs) == 1:
                return xs[0]
            else:
                return xs[0] and reduce(xs[1:])
        return reduce([numpy.all(x) for x in awkward1._util.completely_flatten(layout)])
    else:
        behavior = awkward1._util.behaviorof(array)
        return awkward1._util.wrap(layout.all(axis=axis, mask=False, keepdims=keepdims), behavior)

@awkward1._numpy.implements(numpy.min)
def min(array, axis=None, keepdims=False):
    layout = awkward1.operations.convert.tolayout(array, allowrecord=False, allowother=False)
    if axis is None:
        def reduce(xs):
            if len(xs) == 0:
                return None
            elif len(xs) == 1:
                return xs[0]
            else:
                x, y = xs[0], reduce(xs[1:])
                return x if x < y else y
        tmp = awkward1._util.completely_flatten(layout)
        return reduce([numpy.min(x) for x in tmp if len(x) > 0])
    else:
        behavior = awkward1._util.behaviorof(array)
        return awkward1._util.wrap(layout.min(axis=axis, mask=True, keepdims=keepdims), behavior)

@awkward1._numpy.implements(numpy.max)
def max(array, axis=None, keepdims=False):
    layout = awkward1.operations.convert.tolayout(array, allowrecord=False, allowother=False)
    if axis is None:
        def reduce(xs):
            if len(xs) == 0:
                return None
            elif len(xs) == 1:
                return xs[0]
            else:
                x, y = xs[0], reduce(xs[1:])
                return x if x > y else y
        tmp = awkward1._util.completely_flatten(layout)
        return reduce([numpy.max(x) for x in tmp if len(x) > 0])
    else:
        behavior = awkward1._util.behaviorof(array)
        return awkward1._util.wrap(layout.max(axis=axis, mask=True, keepdims=keepdims), behavior)

### The following are not strictly reducers, but are defined in terms of reducers and ufuncs.

def moment(x, n, weight=None, axis=None, keepdims=False):
    with numpy.errstate(invalid="ignore"):
        if weight is None:
            sumw   = count(x, axis=axis, keepdims=keepdims)
            sumwxn = sum(x**n, axis=axis, keepdims=keepdims)
        else:
            sumw   = sum(x*0 + weight, axis=axis, keepdims=keepdims)
            sumwxn = sum((x*weight)**n, axis=axis, keepdims=keepdims)
        return numpy.true_divide(sumwxn, sumw)

@awkward1._numpy.implements(numpy.mean)
def mean(x, weight=None, axis=None, keepdims=False):
    with numpy.errstate(invalid="ignore"):
        if weight is None:
            sumw  = count(x, axis=axis, keepdims=keepdims)
            sumwx = sum(x, axis=axis, keepdims=keepdims)
        else:
            sumw  = sum(x*0 + weight, axis=axis, keepdims=keepdims)
            sumwx = sum(x*weight, axis=axis, keepdims=keepdims)
        return numpy.true_divide(sumwx, sumw)

@awkward1._numpy.implements(numpy.var)
def var(x, weight=None, ddof=0, axis=None, keepdims=False):
    with numpy.errstate(invalid="ignore"):
        xmean = mean(x, weight=weight, axis=axis, keepdims=keepdims)
        if weight is None:
            sumw   = count(x, axis=axis, keepdims=keepdims)
            sumwxx = sum((x - xmean)**2, axis=axis, keepdims=keepdims)
        else:
            sumw   = sum(x*0 + weight, axis=axis, keepdims=keepdims)
            sumwxx = sum((x - xmean)**2 * weight, axis=axis, keepdims=keepdims)
        if ddof != 0:
            return numpy.true_divide(sumwxx, sumw) * numpy.true_divide(sumw, sumw - ddof)
        else:
            return numpy.true_divide(sumwxx, sumw)

@awkward1._numpy.implements(numpy.std)
def std(x, weight=None, ddof=0, axis=None, keepdims=False):
    with numpy.errstate(invalid="ignore"):
        return numpy.sqrt(var(x, weight=weight, ddof=ddof, axis=axis, keepdims=keepdims))

def covar(x, y, weight=None, axis=None, keepdims=False):
    with numpy.errstate(invalid="ignore"):
        xmean = mean(x, weight=weight, axis=axis, keepdims=keepdims)
        ymean = mean(y, weight=weight, axis=axis, keepdims=keepdims)
        if weight is None:
            sumw = count(x, axis=axis, keepdims=keepdims)
            sumwxy = sum((x - xmean)*(y - ymean), axis=axis, keepdims=keepdims)
        else:
            sumw = sum(x*0 + weight, axis=axis, keepdims=keepdims)
            sumwxy = sum((x - xmean)*(y - ymean)*weight, axis=axis, keepdims=keepdims)
        return numpy.true_divide(sumwxy, sumw)

def corr(x, y, weight=None, axis=None, keepdims=False):
    with numpy.errstate(invalid="ignore"):
        xmean = mean(x, weight=weight, axis=axis, keepdims=keepdims)
        ymean = mean(y, weight=weight, axis=axis, keepdims=keepdims)
        xdiff = x - xmean
        ydiff = y - ymean
        if weight is None:
            sumwxx = sum(xdiff**2, axis=axis, keepdims=keepdims)
            sumwyy = sum(ydiff**2, axis=axis, keepdims=keepdims)
            sumwxy = sum(xdiff*ydiff, axis=axis, keepdims=keepdims)
        else:
            sumwxx = sum((xdiff**2)*weight, axis=axis, keepdims=keepdims)
            sumwyy = sum((ydiff**2)*weight, axis=axis, keepdims=keepdims)
            sumwxy = sum((xdiff*ydiff)*weight, axis=axis, keepdims=keepdims)
        return numpy.true_divide(sumwxy, numpy.sqrt(sumwxx * sumwyy))

def linearfit(x, y, weight=None, axis=None, keepdims=False):
    with numpy.errstate(invalid="ignore"):
        if weight is None:
            sumw = count(x, axis=axis, keepdims=keepdims)
            sumwx  = sum(x, axis=axis, keepdims=keepdims)
            sumwy  = sum(y, axis=axis, keepdims=keepdims)
            sumwxx = sum(x**2, axis=axis, keepdims=keepdims)
            sumwxy = sum(x*y, axis=axis, keepdims=keepdims)
        else:
            sumw = sum(x*0 + weight, axis=axis, keepdims=keepdims)
            sumwx  = sum(x*weight, axis=axis, keepdims=keepdims)
            sumwy  = sum(y*weight, axis=axis, keepdims=keepdims)
            sumwxx = sum((x**2)*weight, axis=axis, keepdims=keepdims)
            sumwxy = sum(x*y*weight, axis=axis, keepdims=keepdims)
        delta           = (sumw*sumwxx) - (sumwx*sumwx)
        intercept       = numpy.true_divide(((sumwxx*sumwy) - (sumwx*sumwxy)), delta)
        slope           = numpy.true_divide(((sumw*sumwxy) - (sumwx*sumwy)), delta)
        intercept_error = numpy.sqrt(numpy.true_divide(sumwxx, delta))
        slope_error     = numpy.sqrt(numpy.true_divide(sumw, delta))

        intercept       = awkward1.operations.convert.tolayout(intercept, allowrecord=True, allowother=True)
        slope           = awkward1.operations.convert.tolayout(slope, allowrecord=True, allowother=True)
        intercept_error = awkward1.operations.convert.tolayout(intercept_error, allowrecord=True, allowother=True)
        slope_error     = awkward1.operations.convert.tolayout(slope_error, allowrecord=True, allowother=True)

        scalar = not isinstance(intercept, awkward1.layout.Content) and not isinstance(slope, awkward1.layout.Content) and not isinstance(intercept_error, awkward1.layout.Content) and not isinstance(slope_error, awkward1.layout.Content)

        if not isinstance(intercept, (awkward1.layout.Content, awkward1.layout.Record)):
            intercept = awkward1.layout.NumpyArray(numpy.array([intercept]))
        if not isinstance(slope, (awkward1.layout.Content, awkward1.layout.Record)):
            slope = awkward1.layout.NumpyArray(numpy.array([slope]))
        if not isinstance(intercept_error, (awkward1.layout.Content, awkward1.layout.Record)):
            intercept_error = awkward1.layout.NumpyArray(numpy.array([intercept_error]))
        if not isinstance(slope_error, (awkward1.layout.Content, awkward1.layout.Record)):
            slope_error = awkward1.layout.NumpyArray(numpy.array([slope_error]))

        out = awkward1.layout.RecordArray([intercept, slope, intercept_error, slope_error], ["intercept", "slope", "intercept_error", "slope_error"])
        out.setparameter("__record__", "LinearFit")
        if scalar:
            out = out[0]

        return awkward1._util.wrap(out, awkward1._util.behaviorof(x, y))

def softmax(x, axis=None, keepdims=False):
    with numpy.errstate(invalid="ignore"):
        expx = numpy.exp(x)
        denom = sum(expx, axis=axis, keepdims=keepdims)
        return numpy.true_divide(expx, denom)

__all__ = [x for x in list(globals()) if not x.startswith("_") and x not in ("collections", "numpy", "awkward1")]



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import sys

import pytest
import numpy

import awkward1

py27 = (sys.version_info[0] < 3)

def test_parameters_on_arrays():
    a = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3, 4.4, 5.5]))
    assert a.parameters == {}
    a.setparameter("one", ["two", 3, {"four": 5}])
    assert a.parameters == {"one": ["two", 3, {"four": 5}]}

    b = awkward1.layout.ListOffsetArray64(awkward1.layout.Index64(numpy.array([0, 3, 3, 5], dtype=numpy.int64)), a)
    assert b.parameters == {}
    assert b.content.parameters == {"one": ["two", 3, {"four": 5}]}
    b.setparameter("what", "ever")
    assert b.parameters == {"what": "ever"}
    assert b.content.parameters == {"one": ["two", 3, {"four": 5}]}

def test_string2():
    content = awkward1.layout.NumpyArray(numpy.array([ord(x) for x in "heythere"], dtype=numpy.uint8), parameters=awkward1.utf8.parameters)
    listoffsetarray = awkward1.layout.ListOffsetArray64(awkward1.layout.Index64(numpy.array([0, 3, 3, 8])), content, parameters=awkward1.string.parameters)
    a = awkward1.Array(listoffsetarray)

    assert isinstance(a, awkward1.Array)
    assert awkward1.tolist(a) == ['hey', '', 'there']

    assert repr(a.type) == "3 * string"
    assert repr(a[0].type) == "3 * utf8"
    assert repr(a[1].type) == "0 * utf8"
    assert repr(a[2].type) == "5 * utf8"

    if py27:
        assert repr(a) == "<Array [u'hey', u'', u'there'] type='3 * string'>"
        assert repr(a[0]) == "u'hey'"
        assert repr(a[1]) == "u''"
        assert repr(a[2]) == "u'there'"
    else:
        assert repr(a) == "<Array ['hey', '', 'there'] type='3 * string'>"
        assert repr(a[0]) == "'hey'"
        assert repr(a[1]) == "''"
        assert repr(a[2]) == "'there'"

def test_dress():
    class Dummy(awkward1.highlevel.Array):
        def __repr__(self):
            return "<Dummy {0}>".format(str(self))
    ns = {"Dummy": Dummy}

    x = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3, 4.4, 5.5]))
    x.setparameter("__array__", "Dummy")
    x.setparameter("__typestr__", "D[5 * float64]")
    a = awkward1.Array(x, behavior=ns)
    assert repr(a) == "<Dummy [1.1, 2.2, 3.3, 4.4, 5.5]>"

    x2 = awkward1.layout.ListOffsetArray64(awkward1.layout.Index64(numpy.array([0, 3, 3, 5], dtype=numpy.int64)), awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3, 4.4, 5.5]), parameters={"__array__": "Dummy"}))
    a2 = awkward1.Array(x2, behavior=ns)
    assert repr(a2) == "<Array [<Dummy [1.1, 2.2, 3.3]>, ... ] type='3 * var * float64[parameters={\"__ar...'>"
    assert repr(a2[0]) == "<Dummy [1.1, 2.2, 3.3]>"
    assert repr(a2[1]) == "<Dummy []>"
    assert repr(a2[2]) == "<Dummy [4.4, 5.5]>"



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import sys

import pytest
import numpy

import awkward1

primes = [x for x in range(2, 1000) if all(x % n != 0 for n in range(2, x))]

def test_reproduce_numpy():
    content1 = awkward1.layout.NumpyArray(numpy.array(primes[:2*3*5], dtype=numpy.int64))
    offsets1 = awkward1.layout.Index64(numpy.array([0, 5, 10, 15, 20, 25, 30], dtype=numpy.int64))
    offsets2 = awkward1.layout.Index64(numpy.array([0, 3, 6], dtype=numpy.int64))
    depth2 = awkward1.layout.ListOffsetArray64(offsets2, awkward1.layout.ListOffsetArray64(offsets1, content1))
    assert awkward1.tolist(depth2) == [
        [[  2,   3,   5,   7,  11],
         [ 13,  17,  19,  23,  29],
         [ 31,  37,  41,  43,  47]],
        [[ 53,  59,  61,  67,  71],
         [ 73,  79,  83,  89,  97],
         [101, 103, 107, 109, 113]]]

    assert awkward1.tolist(depth2.prod(axis=-1)) == [
         [  2 *   3 *   5 *   7 *  11,
           13 *  17 *  19 *  23 *  29,
           31 *  37 *  41 *  43 *  47],
         [ 53 *  59 *  61 *  67 *  71,
           73 *  79 *  83 *  89 *  97,
          101 * 103 * 107 * 109 * 113]]
    assert awkward1.tolist(depth2.prod(axis=2)) == [
         [  2 *   3 *   5 *   7 *  11,
           13 *  17 *  19 *  23 *  29,
           31 *  37 *  41 *  43 *  47],
         [ 53 *  59 *  61 *  67 *  71,
           73 *  79 *  83 *  89 *  97,
          101 * 103 * 107 * 109 * 113]]

    assert awkward1.tolist(depth2.prod(axis=-2)) == [
        [2*13*31, 3*17*37, 5*19*41, 7*23*43, 11*29*47],
        [53*73*101, 59*79*103, 61*83*107, 67*89*109, 71*97*113]]
    assert awkward1.tolist(depth2.prod(axis=1)) == [
        [2*13*31, 3*17*37, 5*19*41, 7*23*43, 11*29*47],
        [53*73*101, 59*79*103, 61*83*107, 67*89*109, 71*97*113]]

    assert awkward1.tolist(depth2.prod(axis=-3)) == [
        [2*53, 3*59, 5*61, 7*67, 11*71],
        [13*73, 17*79, 19*83, 23*89, 29*97],
        [31*101, 37*103, 41*107, 43*109, 47*113]]
    assert awkward1.tolist(depth2.prod(axis=0)) == [
        [2*53, 3*59, 5*61, 7*67, 11*71],
        [13*73, 17*79, 19*83, 23*89, 29*97],
        [31*101, 37*103, 41*107, 43*109, 47*113]]

    content2 = awkward1.layout.NumpyArray(numpy.array(primes[:12], dtype=numpy.int64))
    offsets3 = awkward1.layout.Index64(numpy.array([0, 4, 8, 12], dtype=numpy.int64))
    depth1 = awkward1.layout.ListOffsetArray64(offsets3, content2)

    assert awkward1.tolist(depth1.prod(-1)) == [
        2*3*5*7,
        11*13*17*19,
        23*29*31*37]
    assert awkward1.tolist(depth1.prod(1)) == [
        2*3*5*7,
        11*13*17*19,
        23*29*31*37]

    assert awkward1.tolist(depth1.prod(-2)) == [
        2*11*23,
        3*13*29,
        5*17*31,
        7*19*37]
    assert awkward1.tolist(depth1.prod(0)) == [
        2*11*23,
        3*13*29,
        5*17*31,
        7*19*37]

def test_gaps():
    content1 = awkward1.layout.NumpyArray(numpy.array([123] + primes[:2*3*5], dtype=numpy.int64))
    offsets1 = awkward1.layout.Index64(numpy.array([0, 1, 6, 11, 16, 21, 26, 31], dtype=numpy.int64))
    offsets2 = awkward1.layout.Index64(numpy.array([1, 4, 7], dtype=numpy.int64))
    depth2 = awkward1.layout.ListOffsetArray64(offsets2, awkward1.layout.ListOffsetArray64(offsets1, content1))
    assert awkward1.tolist(depth2) == [
        [[  2,   3,   5,   7,  11],
         [ 13,  17,  19,  23,  29],
         [ 31,  37,  41,  43,  47]],
        [[ 53,  59,  61,  67,  71],
         [ 73,  79,  83,  89,  97],
         [101, 103, 107, 109, 113]]]

    assert awkward1.tolist(depth2.prod(-3)) == [
        [ 106,  177,  305,  469,  781],
        [ 949, 1343, 1577, 2047, 2813],
        [3131, 3811, 4387, 4687, 5311]]

    content1 = awkward1.layout.NumpyArray(numpy.array(primes[:2*3*5 - 1], dtype=numpy.int64))
    offsets1 = awkward1.layout.Index64(numpy.array([0, 5, 10, 15, 20, 25, 29], dtype=numpy.int64))
    offsets2 = awkward1.layout.Index64(numpy.array([0, 3, 6], dtype=numpy.int64))
    depth2 = awkward1.layout.ListOffsetArray64(offsets2, awkward1.layout.ListOffsetArray64(offsets1, content1))
    assert awkward1.tolist(depth2) == [
        [[  2,   3,   5,   7,  11],
         [ 13,  17,  19,  23,  29],
         [ 31,  37,  41,  43,  47]],
        [[ 53,  59,  61,  67,  71],
         [ 73,  79,  83,  89,  97],
         [101, 103, 107, 109,    ]]]

    assert awkward1.tolist(depth2.prod(-3)) == [
        [ 106,  177,  305,  469,  781],
        [ 949, 1343, 1577, 2047, 2813],
        [3131, 3811, 4387, 4687,   47]]

    content1 = awkward1.layout.NumpyArray(numpy.array(primes[:2*3*5 - 2], dtype=numpy.int64))
    offsets1 = awkward1.layout.Index64(numpy.array([0, 5, 10, 15, 20, 25, 28], dtype=numpy.int64))
    offsets2 = awkward1.layout.Index64(numpy.array([0, 3, 6], dtype=numpy.int64))
    depth2 = awkward1.layout.ListOffsetArray64(offsets2, awkward1.layout.ListOffsetArray64(offsets1, content1))
    assert awkward1.tolist(depth2) == [
        [[  2,   3,   5,   7,  11],
         [ 13,  17,  19,  23,  29],
         [ 31,  37,  41,  43,  47]],
        [[ 53,  59,  61,  67,  71],
         [ 73,  79,  83,  89,  97],
         [101, 103, 107,         ]]]

    assert awkward1.tolist(depth2.prod(-3)) == [
        [ 106,  177,  305,  469,  781],
        [ 949, 1343, 1577, 2047, 2813],
        [3131, 3811, 4387,   43,   47]]

    content1 = awkward1.layout.NumpyArray(numpy.array([2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 101, 103, 107, 109], dtype=numpy.int64))
    offsets1 = awkward1.layout.Index64(numpy.array([0, 5, 10, 15, 20, 24, 28], dtype=numpy.int64))
    offsets2 = awkward1.layout.Index64(numpy.array([0, 3, 6], dtype=numpy.int64))
    depth2 = awkward1.layout.ListOffsetArray64(offsets2, awkward1.layout.ListOffsetArray64(offsets1, content1))
    assert awkward1.tolist(depth2) == [
        [[  2,   3,   5,   7,  11],
         [ 13,  17,  19,  23,  29],
         [ 31,  37,  41,  43,  47]],
        [[ 53,  59,  61,  67,  71],
         [ 73,  79,  83,  89,    ],
         [101, 103, 107, 109     ]]]

    assert awkward1.tolist(depth2.prod(-3)) == [
        [ 106,  177,  305,  469, 781],
        [ 949, 1343, 1577, 2047,  29],
        [3131, 3811, 4387, 4687,  47]]

    content1 = awkward1.layout.NumpyArray(numpy.array(primes[1:2*3*5], dtype=numpy.int64))
    offsets1 = awkward1.layout.Index64(numpy.array([0, 4, 9, 14, 19, 24, 29], dtype=numpy.int64))
    offsets2 = awkward1.layout.Index64(numpy.array([0, 3, 6], dtype=numpy.int64))
    depth2 = awkward1.layout.ListOffsetArray64(offsets2, awkward1.layout.ListOffsetArray64(offsets1, content1))
    assert awkward1.tolist(depth2) == [
        [[  3,   5,   7,  11     ],
         [ 13,  17,  19,  23,  29],
         [ 31,  37,  41,  43,  47]],
        [[ 53,  59,  61,  67,  71],
         [ 73,  79,  83,  89,  97],
         [101, 103, 107, 109, 113]]]

    assert awkward1.tolist(depth2.prod(-3)) == [
        [ 159,  295,  427,  737,   71],
        [ 949, 1343, 1577, 2047, 2813],
        [3131, 3811, 4387, 4687, 5311]]

    content1 = awkward1.layout.NumpyArray(numpy.array(primes[2:2*3*5], dtype=numpy.int64))
    offsets1 = awkward1.layout.Index64(numpy.array([0, 3, 8, 13, 18, 23, 28], dtype=numpy.int64))
    offsets2 = awkward1.layout.Index64(numpy.array([0, 3, 6], dtype=numpy.int64))
    depth2 = awkward1.layout.ListOffsetArray64(offsets2, awkward1.layout.ListOffsetArray64(offsets1, content1))
    assert awkward1.tolist(depth2) == [
        [[  5,   7,  11          ],
         [ 13,  17,  19,  23,  29],
         [ 31,  37,  41,  43,  47]],
        [[ 53,  59,  61,  67,  71],
         [ 73,  79,  83,  89,  97],
         [101, 103, 107, 109, 113]]]

    assert awkward1.tolist(depth2.prod(-3)) == [
        [ 265,  413,  671,   67,   71],
        [ 949, 1343, 1577, 2047, 2813],
        [3131, 3811, 4387, 4687, 5311]]

    content1 = awkward1.layout.NumpyArray(numpy.array([3, 5, 7, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97, 101, 103, 107, 109, 113], dtype=numpy.int64))
    offsets1 = awkward1.layout.Index64(numpy.array([0, 3, 8, 13, 18, 23, 28], dtype=numpy.int64))
    offsets2 = awkward1.layout.Index64(numpy.array([0, 3, 6], dtype=numpy.int64))
    depth2 = awkward1.layout.ListOffsetArray64(offsets2, awkward1.layout.ListOffsetArray64(offsets1, content1))
    assert awkward1.tolist(depth2) == [
        [[  3,   5,   7,         ],
         [ 13,  17,  19,  23,  29],
         [ 31,  37,  41,  43,  47]],
        [[ 53,  59,  61,  67,  71],
         [ 73,  79,  83,  89,  97],
         [101, 103, 107, 109, 113]]]

    assert awkward1.tolist(depth2.prod(-3)) == [
        [ 159,  295,  427,   67,   71],
        [ 949, 1343, 1577, 2047, 2813],
        [3131, 3811, 4387, 4687, 5311]]

    content1 = awkward1.layout.NumpyArray(numpy.array([3, 5, 7, 11, 13, 17, 19, 23, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97, 101, 103, 107, 109, 113], dtype=numpy.int64))
    offsets1 = awkward1.layout.Index64(numpy.array([0, 4, 8, 13, 18, 23, 28], dtype=numpy.int64))
    offsets2 = awkward1.layout.Index64(numpy.array([0, 3, 6], dtype=numpy.int64))
    depth2 = awkward1.layout.ListOffsetArray64(offsets2, awkward1.layout.ListOffsetArray64(offsets1, content1))
    assert awkward1.tolist(depth2) == [
        [[  3,   5,   7,  11     ],
         [ 13,  17,  19,  23     ],
         [ 31,  37,  41,  43,  47]],
        [[ 53,  59,  61,  67,  71],
         [ 73,  79,  83,  89,  97],
         [101, 103, 107, 109, 113]]]

    assert awkward1.tolist(depth2.prod(-3)) == [
        [ 159,  295,  427,  737,   71],
        [ 949, 1343, 1577, 2047,   97],
        [3131, 3811, 4387, 4687, 5311]]

    content1 = awkward1.layout.NumpyArray(numpy.array([2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97, 101, 103, 107, 109], dtype=numpy.int64))
    offsets1 = awkward1.layout.Index64(numpy.array([0, 5, 10, 14, 19, 24, 28], dtype=numpy.int64))
    offsets2 = awkward1.layout.Index64(numpy.array([0, 3, 6], dtype=numpy.int64))
    depth2 = awkward1.layout.ListOffsetArray64(offsets2, awkward1.layout.ListOffsetArray64(offsets1, content1))
    assert awkward1.tolist(depth2) == [
        [[  2,   3,   5,   7,  11],
         [ 13,  17,  19,  23,  29],
         [ 31,  37,  41,  43     ]],
        [[ 53,  59,  61,  67,  71],
         [ 73,  79,  83,  89,  97],
         [101, 103, 107, 109     ]]]

    assert awkward1.tolist(depth2.prod(-3)) == [
        [ 106,  177,  305,  469,  781],
        [ 949, 1343, 1577, 2047, 2813],
        [3131, 3811, 4387, 4687]]

    content1 = awkward1.layout.NumpyArray(numpy.array([2, 3, 5, 7, 11, 13, 17, 19, 23, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 101, 103, 107, 109, 113], dtype=numpy.int64))
    offsets1 = awkward1.layout.Index64(numpy.array([0, 5, 9, 14, 19, 23, 28], dtype=numpy.int64))
    offsets2 = awkward1.layout.Index64(numpy.array([0, 3, 6], dtype=numpy.int64))
    depth2 = awkward1.layout.ListOffsetArray64(offsets2, awkward1.layout.ListOffsetArray64(offsets1, content1))
    assert awkward1.tolist(depth2) == [
        [[  2,   3,   5,   7,  11],
         [ 13,  17,  19,  23     ],
         [ 31,  37,  41,  43,  47]],
        [[ 53,  59,  61,  67,  71],
         [ 73,  79,  83,  89     ],
         [101, 103, 107, 109, 113]]]

    assert awkward1.tolist(depth2.prod(-3)) == [
        [ 106,  177,  305,  469,  781],
        [ 949, 1343, 1577, 2047      ],
        [3131, 3811, 4387, 4687, 5311]]

    content1 = awkward1.layout.NumpyArray(numpy.array(primes[:9], dtype=numpy.int64))
    offsets1 = awkward1.layout.Index64(numpy.array([0, 3, 4, 6, 6, 7, 9], dtype=numpy.int64))
    offsets2 = awkward1.layout.Index64(numpy.array([0, 2, 4, 6], dtype=numpy.int64))
    depth2 = awkward1.layout.ListOffsetArray64(offsets2, awkward1.layout.ListOffsetArray64(offsets1, content1))
    assert awkward1.tolist(depth2) == [
        [[      2,   3, 5],
         [      7        ]],
        [[     11,  13   ],
         [               ]],
        [[     17        ],
         [     19,  23   ]]]

    assert awkward1.tolist(depth2.prod(-3)) == [
        [2*11*17, 3*13, 5],
        [7*19   , 23     ]]

    content1 = awkward1.layout.NumpyArray(numpy.array(primes[:9], dtype=numpy.int64))
    offsets1 = awkward1.layout.Index64(numpy.array([0, 3, 4, 6, 7, 9], dtype=numpy.int64))
    offsets2 = awkward1.layout.Index64(numpy.array([0, 2, 3, 5], dtype=numpy.int64))
    depth2 = awkward1.layout.ListOffsetArray64(offsets2, awkward1.layout.ListOffsetArray64(offsets1, content1))
    assert awkward1.tolist(depth2) == [
        [[      2,   3, 5],
         [      7        ]],
        [[     11,  13   ]],
        [[     17        ],
         [     19,  23   ]]]

    assert awkward1.tolist(depth2.prod(-3)) == [
        [2*11*17, 3*13, 5],
        [7*19   , 23     ]]

    content1 = awkward1.layout.NumpyArray(numpy.array(primes[:10], dtype=numpy.int64))
    offsets1 = awkward1.layout.Index64(numpy.array([0, 3, 5, 6, 8, 9, 10], dtype=numpy.int64))
    offsets2 = awkward1.layout.Index64(numpy.array([0, 3, 6], dtype=numpy.int64))
    depth2 = awkward1.layout.ListOffsetArray64(offsets2, awkward1.layout.ListOffsetArray64(offsets1, content1))
    assert awkward1.tolist(depth2) == [
        [[ 2,  3, 5],
         [ 7, 11   ],
         [13       ]],
        [[17, 19   ],
         [23       ],
         [29       ]]]

    assert awkward1.tolist(depth2.prod(-3)) == [
        [ 34, 57, 5],
        [161, 11   ],
        [377       ]]

    content1 = awkward1.layout.NumpyArray(numpy.array(primes[:9], dtype=numpy.int64))
    offsets1 = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6, 8, 9], dtype=numpy.int64))
    offsets2 = awkward1.layout.Index64(numpy.array([0, 4, 6], dtype=numpy.int64))
    depth2 = awkward1.layout.ListOffsetArray64(offsets2, awkward1.layout.ListOffsetArray64(offsets1, content1))
    assert awkward1.tolist(depth2) == [
        [[ 2,  3, 5],
         [         ],
         [ 7, 11   ],
         [13       ]],
        [[17, 19   ],
         [23       ]]]

    assert awkward1.tolist(depth2.prod(-3)) == [
        [34, 57, 5],
        [23       ],
        [ 7, 11   ],
        [13       ]]

    content1 = awkward1.layout.NumpyArray(numpy.array(primes[:9], dtype=numpy.int64))
    offsets1 = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6, 8, 9], dtype=numpy.int64))
    offsets2 = awkward1.layout.Index64(numpy.array([0, 4, 4, 6], dtype=numpy.int64))
    depth2 = awkward1.layout.ListOffsetArray64(offsets2, awkward1.layout.ListOffsetArray64(offsets1, content1))
    assert awkward1.tolist(depth2) == [
        [[ 2,  3, 5],
         [         ],
         [ 7, 11   ],
         [13       ]],
        [],
        [[17, 19   ],
         [23       ]]]

    assert awkward1.tolist(depth2.prod(-3)) == [
        [34, 57, 5],
        [23       ],
        [ 7, 11   ],
        [13       ]]

    content1 = awkward1.layout.NumpyArray(numpy.array(primes[:2*3*5], dtype=numpy.int64))
    offsets1 = awkward1.layout.Index64(numpy.array([0, 5, 10, 15, 20, 25, 30], dtype=numpy.int64))
    offsets2 = awkward1.layout.Index64(numpy.array([0, 3, 6], dtype=numpy.int64))
    depth2 = awkward1.layout.ListOffsetArray64(offsets2, awkward1.layout.ListOffsetArray64(offsets1, content1))
    assert awkward1.tolist(depth2) == [
        [[  2,   3,   5,   7,  11],
         [ 13,  17,  19,  23,  29],
         [ 31,  37,  41,  43,  47]],
        [[ 53,  59,  61,  67,  71],
         [ 73,  79,  83,  89,  97],
         [101, 103, 107, 109, 113]]]

    assert awkward1.tolist(depth2.prod(-1)) == [
        [  2 *   3 *   5 *   7 *  11,
          13 *  17 *  19 *  23 *  29,
          31 *  37 *  41 *  43 *  47],
        [ 53 *  59 *  61 *  67 *  71,
          73 *  79 *  83 *  89 *  97,
         101 * 103 * 107 * 109 * 113]]

    assert awkward1.tolist(depth2.prod(-2)) == [
        [2*13*31, 3*17*37, 5*19*41, 7*23*43, 11*29*47],
        [53*73*101, 59*79*103, 61*83*107, 67*89*109, 71*97*113]]

    content1 = awkward1.layout.NumpyArray(numpy.array(primes[:9], dtype=numpy.int64))
    offsets1 = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6, 8, 9], dtype=numpy.int64))
    offsets2 = awkward1.layout.Index64(numpy.array([0, 4, 4, 6], dtype=numpy.int64))
    depth2 = awkward1.layout.ListOffsetArray64(offsets2, awkward1.layout.ListOffsetArray64(offsets1, content1))
    assert awkward1.tolist(depth2) == [
        [[ 2,  3, 5],
         [         ],
         [ 7, 11   ],
         [13       ]],
        [],
        [[17, 19   ],
         [23       ]]]

    assert awkward1.tolist(depth2.prod(-1)) == [
        [2*3*5, 1, 7*11, 13],
        [],
        [17*19, 23]]

    assert awkward1.tolist(depth2.prod(-2)) == [
        [2*7*13, 3*11, 5],
        [],
        [17*23, 19]]

def test_complicated():
    offsets1 = awkward1.layout.Index64(numpy.array([0, 3, 3, 5], dtype=numpy.int64))
    content1 = awkward1.layout.ListOffsetArray64(offsets1, awkward1.layout.NumpyArray(numpy.array(primes[:5], dtype=numpy.int64)))
    offsets2 = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6, 8, 9], dtype=numpy.int64))
    offsets3 = awkward1.layout.Index64(numpy.array([0, 4, 4, 6], dtype=numpy.int64))
    content2 = awkward1.layout.ListOffsetArray64(offsets3, awkward1.layout.ListOffsetArray64(offsets2, awkward1.layout.NumpyArray(numpy.array(primes[:9], dtype=numpy.int64))))
    offsets4 = awkward1.layout.Index64(numpy.array([0, 1, 1, 3], dtype=numpy.int64))
    complicated = awkward1.layout.ListOffsetArray64(offsets4, awkward1.layout.RecordArray([content1, content2], ["x", "y"]))

    assert awkward1.tolist(complicated) == [[{"x": [2, 3, 5], "y": [[2, 3, 5], [], [7, 11], [13]]}], [], [{"x": [], "y": []}, {"x": [7, 11], "y": [[17, 19], [23]]}]]

    assert awkward1.tolist(complicated["x"]) == [
        [[2, 3, 5]],
        [],
        [[],
         [7, 11]]]
    assert awkward1.tolist(complicated["y"]) == [
        [[[ 2,  3, 5],
          [         ],
          [ 7, 11   ],
          [13       ]]],
        [             ],
        [[          ],
         [[17, 19   ],
          [23       ]]]]

    assert awkward1.tolist(complicated.prod(-1)) == [{"x": [30], "y": [[30, 1, 77, 13]]}, {"x": [], "y": []}, {"x": [1, 77], "y": [[], [323, 23]]}]
    assert awkward1.tolist(complicated["x"].prod(-1)) == [[30], [], [1, 77]]
    assert awkward1.tolist(complicated["y"].prod(-1)) == [[[30, 1, 77, 13]], [], [[], [323, 23]]]

    assert awkward1.tolist(complicated.prod(-2)) == [{"x": [2, 3, 5], "y": [[182, 33, 5]]}, {"x": [], "y": []}, {"x": [7, 11], "y": [[], [391, 19]]}]
    assert awkward1.tolist(complicated["x"].prod(-2)) == [[2, 3, 5], [], [7, 11]]
    assert awkward1.tolist(complicated["y"].prod(-2)) == [[[182, 33, 5]], [], [[], [391, 19]]]

    assert awkward1.tolist(complicated[0]) == [{"x": [2, 3, 5], "y": [[2, 3, 5], [], [7, 11], [13]]}]
    assert awkward1.tolist(complicated[0].prod(-1)) == {"x": [30], "y": [[30, 1, 77, 13]]}

def test_EmptyArray():
    offsets = awkward1.layout.Index64(numpy.array([0, 0, 0, 0], dtype=numpy.int64))
    array = awkward1.layout.ListOffsetArray64(offsets, awkward1.layout.EmptyArray())
    assert awkward1.tolist(array) == [[], [], []]

    assert awkward1.tolist(array.prod(-1)) == [1, 1, 1]

    offsets = awkward1.layout.Index64(numpy.array([0, 0, 0, 0], dtype=numpy.int64))
    array = awkward1.layout.ListOffsetArray64(offsets, awkward1.layout.NumpyArray(numpy.array([], dtype=numpy.int64)))
    assert awkward1.tolist(array) == [[], [], []]

    assert awkward1.tolist(array.prod(-1)) == [1, 1, 1]

def test_IndexedOptionArray():
    content = awkward1.layout.NumpyArray(numpy.array(primes[:2*3*5], dtype=numpy.int64))
    offsets1 = awkward1.layout.Index64(numpy.array([0, 5, 10, 15, 20, 25, 30], dtype=numpy.int64))
    listoffsetarray = awkward1.layout.ListOffsetArray64(offsets1, content)
    index = awkward1.layout.Index64(numpy.array([5, 4, 3, 2, 1, 0], dtype=numpy.int64))
    indexedarray = awkward1.layout.IndexedArray64(index, listoffsetarray)
    offsets2 = awkward1.layout.Index64(numpy.array([0, 3, 6], dtype=numpy.int64))
    depth2 = awkward1.layout.ListOffsetArray64(offsets2, indexedarray)
    assert awkward1.tolist(depth2) == [
        [[101, 103, 107, 109, 113],
         [ 73,  79,  83,  89,  97],
         [ 53,  59,  61,  67,  71]],
        [[ 31,  37,  41,  43,  47],
         [ 13,  17,  19,  23,  29],
         [  2,   3,   5,   7,  11]]]

    assert awkward1.tolist(depth2.prod(-1)) == [
        [101 * 103 * 107 * 109 * 113,
          73 *  79 *  83 *  89 *  97,
          53 *  59 *  61 *  67 *  71],
        [ 31 *  37 *  41 *  43 *  47,
          13 *  17 *  19 *  23 *  29,
           2 *   3 *   5 *   7 *  11]]

    assert awkward1.tolist(depth2.prod(-2)) == [
        [101*73*53, 103*79*59, 107*83*61, 109*89*67, 113*97*71],
        [  31*13*2,   37*17*3,   41*19*5,   43*23*7,  47*29*11]]

    assert awkward1.tolist(depth2.prod(-3)) == [
        [101*31, 103*37, 107*41, 109*43, 113*47],
        [ 73*13,  79*17,  83*19,  89*23,  97*29],
        [  53*2,   59*3,   61*5,   67*7,  71*11]]

    content = awkward1.layout.NumpyArray(numpy.array([2, 3, 5, 7, 11, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 101, 103, 107, 109, 113], dtype=numpy.int64))
    offsets1 = awkward1.layout.Index64(numpy.array([0, 5, 10, 15, 20], dtype=numpy.int64))
    listoffsetarray = awkward1.layout.ListOffsetArray64(offsets1, content)
    index = awkward1.layout.Index64(numpy.array([3, -1, 2, 1, -1, 0], dtype=numpy.int64))
    indexedoptionarray = awkward1.layout.IndexedOptionArray64(index, listoffsetarray)
    offsets2 = awkward1.layout.Index64(numpy.array([0, 3, 6], dtype=numpy.int64))
    depth2 = awkward1.layout.ListOffsetArray64(offsets2, indexedoptionarray)
    assert awkward1.tolist(depth2) == [
        [[101, 103, 107, 109, 113],
         None,
         [ 53,  59,  61,  67,  71]],
        [[ 31,  37,  41,  43,  47],
         None,
         [  2,   3,   5,   7,  11]]]

    assert awkward1.tolist(depth2.prod(-1)) == [
        [101 * 103 * 107 * 109 * 113,
          53 *  59 *  61 *  67 *  71],
        [ 31 *  37 *  41 *  43 *  47,
           2 *   3 *   5 *   7 *  11]]

    assert awkward1.tolist(depth2.prod(-2)) == [
        [101*53, 103*59, 107*61, 109*67, 113*71],
        [  31*2,   37*3,   41*5,   43*7,  47*11]]

    assert awkward1.tolist(depth2.prod(-3)) == [
        [101*31, 103*37, 107*41, 109*43, 113*47],
        [],
        [  53*2,   59*3,   61*5,   67*7,  71*11]]

    content = awkward1.layout.NumpyArray(numpy.array([2, 3, 5, 7, 11, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 101, 103, 107, 109, 113], dtype=numpy.int64))
    index = awkward1.layout.Index64(numpy.array([15, 16, 17, 18, 19, -1, -1, -1, -1, -1, 10, 11, 12, 13, 14, 5, 6, 7, 8, 9, -1, -1, -1, -1, -1, 0, 1, 2, 3, 4], dtype=numpy.int64))
    indexedoptionarray = awkward1.layout.IndexedOptionArray64(index, content)
    offsets1 = awkward1.layout.Index64(numpy.array([0, 5, 10, 15, 20, 25, 30], dtype=numpy.int64))
    listoffsetarray = awkward1.layout.ListOffsetArray64(offsets1, indexedoptionarray)
    offsets2 = awkward1.layout.Index64(numpy.array([0, 3, 6], dtype=numpy.int64))
    depth2 = awkward1.layout.ListOffsetArray64(offsets2, listoffsetarray)
    assert awkward1.tolist(depth2) == [
        [[ 101,  103,  107,  109,  113],
         [None, None, None, None, None],
         [  53,   59,   61,   67,   71]],
        [[  31,   37,   41,   43,   47],
         [None, None, None, None, None],
         [   2,    3,    5,    7,   11]]]

    assert awkward1.tolist(depth2.prod(-1)) == [
        [101 * 103 * 107 * 109 * 113,
           1 *   1 *   1 *   1 *   1,
          53 *  59 *  61 *  67 *  71],
        [ 31 *  37 *  41 *  43 *  47,
           1 *   1 *   1 *   1 *   1,
           2 *   3 *   5 *   7 *  11]]

    assert awkward1.tolist(depth2.prod(-2)) == [
        [101*53, 103*59, 107*61, 109*67, 113*71],
        [  31*2,   37*3,   41*5,   43*7,  47*11]]

    assert awkward1.tolist(depth2.prod(-3)) == [
        [101*31, 103*37, 107*41, 109*43, 113*47],
        [     1,      1,      1,      1,      1],
        [  53*2,   59*3,   61*5,   67*7,  71*11]]

    content = awkward1.layout.NumpyArray(numpy.array([2, 3, 5, 7, 11, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 101, 103, 107, 109, 113], dtype=numpy.int64))
    index = awkward1.layout.Index64(numpy.array([15, 16, 17, 18, 19, -1, 10, 11, 12, 13, 14, 5, 6, 7, 8, 9, -1, 0, 1, 2, 3, 4], dtype=numpy.int64))
    indexedoptionarray = awkward1.layout.IndexedOptionArray64(index, content)
    offsets1 = awkward1.layout.Index64(numpy.array([0, 5, 6, 11, 16, 17, 22], dtype=numpy.int64))
    listoffsetarray = awkward1.layout.ListOffsetArray64(offsets1, indexedoptionarray)
    offsets2 = awkward1.layout.Index64(numpy.array([0, 3, 6], dtype=numpy.int64))
    depth2 = awkward1.layout.ListOffsetArray64(offsets2, listoffsetarray)
    assert awkward1.tolist(depth2) == [
        [[ 101,  103,  107,  109,  113],
         [None],
         [  53,   59,   61,   67,   71]],
        [[  31,   37,   41,   43,   47],
         [None],
         [   2,    3,    5,    7,   11]]]

    assert awkward1.tolist(depth2.prod(-1)) == [
        [101 * 103 * 107 * 109 * 113,
           1,
          53 *  59 *  61 *  67 *  71],
        [ 31 *  37 *  41 *  43 *  47,
           1,
           2 *   3 *   5 *   7 *  11]]

    assert awkward1.tolist(depth2.prod(-2)) == [
        [101*53, 103*59, 107*61, 109*67, 113*71],
        [  31*2,   37*3,   41*5,   43*7,  47*11]]

    assert awkward1.tolist(depth2.prod(-3)) == [
        [101*31, 103*37, 107*41, 109*43, 113*47],
        [     1],
        [  53*2,   59*3,   61*5,   67*7,  71*11]]

def test_UnionArray():
    content1 = awkward1.Array([
        [[  2,   3,   5,   7,  11],
         [ 13,  17,  19,  23,  29],
         [ 31,  37,  41,  43,  47]]]).layout
    content2 = awkward1.Array([
        [[ 53,  59,  61,  67,  71],
         [ 73,  79,  83,  89,  97],
         [101, 103, 107, 109, 113]]]).layout

    tags = awkward1.layout.Index8(numpy.array([0, 1], dtype=numpy.int8))
    index = awkward1.layout.Index64(numpy.array([0, 0], dtype=numpy.int64))
    depth2 = awkward1.layout.UnionArray8_64(tags, index, [content1, content2])
    assert awkward1.tolist(depth2) == [
        [[  2,   3,   5,   7,  11],
         [ 13,  17,  19,  23,  29],
         [ 31,  37,  41,  43,  47]],
        [[ 53,  59,  61,  67,  71],
         [ 73,  79,  83,  89,  97],
         [101, 103, 107, 109, 113]]]

    assert awkward1.tolist(depth2.prod(axis=-1)) == [
         [  2 *   3 *   5 *   7 *  11,
           13 *  17 *  19 *  23 *  29,
           31 *  37 *  41 *  43 *  47],
         [ 53 *  59 *  61 *  67 *  71,
           73 *  79 *  83 *  89 *  97,
          101 * 103 * 107 * 109 * 113]]
    assert awkward1.tolist(depth2.prod(axis=2)) == [
         [  2 *   3 *   5 *   7 *  11,
           13 *  17 *  19 *  23 *  29,
           31 *  37 *  41 *  43 *  47],
         [ 53 *  59 *  61 *  67 *  71,
           73 *  79 *  83 *  89 *  97,
          101 * 103 * 107 * 109 * 113]]

    assert awkward1.tolist(depth2.prod(axis=-2)) == [
        [2*13*31, 3*17*37, 5*19*41, 7*23*43, 11*29*47],
        [53*73*101, 59*79*103, 61*83*107, 67*89*109, 71*97*113]]
    assert awkward1.tolist(depth2.prod(axis=1)) == [
        [2*13*31, 3*17*37, 5*19*41, 7*23*43, 11*29*47],
        [53*73*101, 59*79*103, 61*83*107, 67*89*109, 71*97*113]]

    assert awkward1.tolist(depth2.prod(axis=-3)) == [
        [2*53, 3*59, 5*61, 7*67, 11*71],
        [13*73, 17*79, 19*83, 23*89, 29*97],
        [31*101, 37*103, 41*107, 43*109, 47*113]]
    assert awkward1.tolist(depth2.prod(axis=0)) == [
        [2*53, 3*59, 5*61, 7*67, 11*71],
        [13*73, 17*79, 19*83, 23*89, 29*97],
        [31*101, 37*103, 41*107, 43*109, 47*113]]

    content1 = awkward1.layout.NumpyArray(numpy.array(primes[:2*3*5], dtype=numpy.int64))
    offsets1a = awkward1.layout.Index64(numpy.array([0, 5, 10, 15], dtype=numpy.int64))
    offsets1b = awkward1.layout.Index64(numpy.array([15, 20, 25, 30], dtype=numpy.int64))
    tags = awkward1.layout.Index8(numpy.array([0, 0, 0, 1, 1, 1], dtype=numpy.int8))
    index = awkward1.layout.Index64(numpy.array([0, 1, 2, 0, 1, 2], dtype=numpy.int64))
    unionarray = awkward1.layout.UnionArray8_64(tags, index, [awkward1.layout.ListOffsetArray64(offsets1a, content1), awkward1.layout.ListOffsetArray64(offsets1b, content1)])
    offsets2 = awkward1.layout.Index64(numpy.array([0, 3, 6], dtype=numpy.int64))
    depth2 = awkward1.layout.ListOffsetArray64(offsets2, unionarray)
    assert awkward1.tolist(depth2) == [
        [[  2,   3,   5,   7,  11],
         [ 13,  17,  19,  23,  29],
         [ 31,  37,  41,  43,  47]],
        [[ 53,  59,  61,  67,  71],
         [ 73,  79,  83,  89,  97],
         [101, 103, 107, 109, 113]]]

    assert awkward1.tolist(depth2.prod(axis=-1)) == [
         [  2 *   3 *   5 *   7 *  11,
           13 *  17 *  19 *  23 *  29,
           31 *  37 *  41 *  43 *  47],
         [ 53 *  59 *  61 *  67 *  71,
           73 *  79 *  83 *  89 *  97,
          101 * 103 * 107 * 109 * 113]]
    assert awkward1.tolist(depth2.prod(axis=2)) == [
         [  2 *   3 *   5 *   7 *  11,
           13 *  17 *  19 *  23 *  29,
           31 *  37 *  41 *  43 *  47],
         [ 53 *  59 *  61 *  67 *  71,
           73 *  79 *  83 *  89 *  97,
          101 * 103 * 107 * 109 * 113]]

    assert awkward1.tolist(depth2.prod(axis=-2)) == [
        [2*13*31, 3*17*37, 5*19*41, 7*23*43, 11*29*47],
        [53*73*101, 59*79*103, 61*83*107, 67*89*109, 71*97*113]]
    assert awkward1.tolist(depth2.prod(axis=1)) == [
        [2*13*31, 3*17*37, 5*19*41, 7*23*43, 11*29*47],
        [53*73*101, 59*79*103, 61*83*107, 67*89*109, 71*97*113]]

    assert awkward1.tolist(depth2.prod(axis=-3)) == [
        [2*53, 3*59, 5*61, 7*67, 11*71],
        [13*73, 17*79, 19*83, 23*89, 29*97],
        [31*101, 37*103, 41*107, 43*109, 47*113]]
    assert awkward1.tolist(depth2.prod(axis=0)) == [
        [2*53, 3*59, 5*61, 7*67, 11*71],
        [13*73, 17*79, 19*83, 23*89, 29*97],
        [31*101, 37*103, 41*107, 43*109, 47*113]]

def test_sum():
    content2 = awkward1.layout.NumpyArray(numpy.array([1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048], dtype=numpy.int64))
    offsets3 = awkward1.layout.Index64(numpy.array([0, 4, 8, 12], dtype=numpy.int64))
    depth1 = awkward1.layout.ListOffsetArray64(offsets3, content2)

    assert awkward1.tolist(depth1.sum(-1)) == [
        1 + 2 + 4 + 8,
        16 + 32 + 64 + 128,
        256 + 512 + 1024 + 2048]
    assert awkward1.tolist(depth1.sum(1)) == [
        1 + 2 + 4 + 8,
        16 + 32 + 64 + 128,
        256 + 512 + 1024 + 2048]

    assert awkward1.tolist(depth1.sum(-2)) == [
        1 + 16 + 256,
        2 + 32 + 512,
        4 + 64 + 1024,
        8 + 128 + 2048]
    assert awkward1.tolist(depth1.sum(0)) == [
        1 + 16 + 256,
        2 + 32 + 512,
        4 + 64 + 1024,
        8 + 128 + 2048]

def test_sumprod_types():
    def prod(xs):
        out = 1
        for x in xs:
            out *= x
        return out

    array = numpy.array([[True, False, False], [True, False, False]])
    content2 = awkward1.layout.NumpyArray(array.reshape(-1))
    offsets3 = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6], dtype=numpy.int64))
    depth1 = awkward1.layout.ListOffsetArray64(offsets3, content2)
    assert numpy.sum(array, axis=-1).dtype == numpy.asarray(depth1.sum(axis=-1)).dtype
    assert numpy.prod(array, axis=-1).dtype == numpy.asarray(depth1.prod(axis=-1)).dtype
    assert sum(awkward1.tolist(numpy.sum(array, axis=-1))) == sum(awkward1.tolist(depth1.sum(axis=-1)))
    assert prod(awkward1.tolist(numpy.prod(array, axis=-1))) == prod(awkward1.tolist(depth1.prod(axis=-1)))

    array = numpy.array([[0, 1, 2], [3, 4, 5]], dtype=numpy.int8)
    content2 = awkward1.layout.NumpyArray(array.reshape(-1))
    offsets3 = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6], dtype=numpy.int64))
    depth1 = awkward1.layout.ListOffsetArray64(offsets3, content2)
    assert numpy.sum(array, axis=-1).dtype == numpy.asarray(depth1.sum(axis=-1)).dtype
    assert numpy.prod(array, axis=-1).dtype == numpy.asarray(depth1.prod(axis=-1)).dtype
    assert sum(awkward1.tolist(numpy.sum(array, axis=-1))) == sum(awkward1.tolist(depth1.sum(axis=-1)))
    assert prod(awkward1.tolist(numpy.prod(array, axis=-1))) == prod(awkward1.tolist(depth1.prod(axis=-1)))

    array = numpy.array([[0, 1, 2], [3, 4, 5]], dtype=numpy.uint8)
    content2 = awkward1.layout.NumpyArray(array.reshape(-1))
    offsets3 = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6], dtype=numpy.int64))
    depth1 = awkward1.layout.ListOffsetArray64(offsets3, content2)
    assert numpy.sum(array, axis=-1).dtype == numpy.asarray(depth1.sum(axis=-1)).dtype
    assert numpy.prod(array, axis=-1).dtype == numpy.asarray(depth1.prod(axis=-1)).dtype
    assert sum(awkward1.tolist(numpy.sum(array, axis=-1))) == sum(awkward1.tolist(depth1.sum(axis=-1)))
    assert prod(awkward1.tolist(numpy.prod(array, axis=-1))) == prod(awkward1.tolist(depth1.prod(axis=-1)))

    array = numpy.array([[0, 1, 2], [3, 4, 5]], dtype=numpy.int16)
    content2 = awkward1.layout.NumpyArray(array.reshape(-1))
    offsets3 = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6], dtype=numpy.int64))
    depth1 = awkward1.layout.ListOffsetArray64(offsets3, content2)
    assert numpy.sum(array, axis=-1).dtype == numpy.asarray(depth1.sum(axis=-1)).dtype
    assert numpy.prod(array, axis=-1).dtype == numpy.asarray(depth1.prod(axis=-1)).dtype
    assert sum(awkward1.tolist(numpy.sum(array, axis=-1))) == sum(awkward1.tolist(depth1.sum(axis=-1)))
    assert prod(awkward1.tolist(numpy.prod(array, axis=-1))) == prod(awkward1.tolist(depth1.prod(axis=-1)))

    array = numpy.array([[0, 1, 2], [3, 4, 5]], dtype=numpy.uint16)
    content2 = awkward1.layout.NumpyArray(array.reshape(-1))
    offsets3 = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6], dtype=numpy.int64))
    depth1 = awkward1.layout.ListOffsetArray64(offsets3, content2)
    assert numpy.sum(array, axis=-1).dtype == numpy.asarray(depth1.sum(axis=-1)).dtype
    assert numpy.prod(array, axis=-1).dtype == numpy.asarray(depth1.prod(axis=-1)).dtype
    assert sum(awkward1.tolist(numpy.sum(array, axis=-1))) == sum(awkward1.tolist(depth1.sum(axis=-1)))
    assert prod(awkward1.tolist(numpy.prod(array, axis=-1))) == prod(awkward1.tolist(depth1.prod(axis=-1)))

    array = numpy.array([[0, 1, 2], [3, 4, 5]], dtype=numpy.int32)
    content2 = awkward1.layout.NumpyArray(array.reshape(-1))
    offsets3 = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6], dtype=numpy.int64))
    depth1 = awkward1.layout.ListOffsetArray64(offsets3, content2)
    assert numpy.sum(array, axis=-1).dtype == numpy.asarray(depth1.sum(axis=-1)).dtype
    assert numpy.prod(array, axis=-1).dtype == numpy.asarray(depth1.prod(axis=-1)).dtype
    assert sum(awkward1.tolist(numpy.sum(array, axis=-1))) == sum(awkward1.tolist(depth1.sum(axis=-1)))
    assert prod(awkward1.tolist(numpy.prod(array, axis=-1))) == prod(awkward1.tolist(depth1.prod(axis=-1)))

    array = numpy.array([[0, 1, 2], [3, 4, 5]], dtype=numpy.uint32)
    content2 = awkward1.layout.NumpyArray(array.reshape(-1))
    offsets3 = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6], dtype=numpy.int64))
    depth1 = awkward1.layout.ListOffsetArray64(offsets3, content2)
    assert numpy.sum(array, axis=-1).dtype == numpy.asarray(depth1.sum(axis=-1)).dtype
    assert numpy.prod(array, axis=-1).dtype == numpy.asarray(depth1.prod(axis=-1)).dtype
    assert sum(awkward1.tolist(numpy.sum(array, axis=-1))) == sum(awkward1.tolist(depth1.sum(axis=-1)))
    assert prod(awkward1.tolist(numpy.prod(array, axis=-1))) == prod(awkward1.tolist(depth1.prod(axis=-1)))

    array = numpy.array([[0, 1, 2], [3, 4, 5]], dtype=numpy.int64)
    content2 = awkward1.layout.NumpyArray(array.reshape(-1))
    offsets3 = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6], dtype=numpy.int64))
    depth1 = awkward1.layout.ListOffsetArray64(offsets3, content2)
    assert numpy.sum(array, axis=-1).dtype == numpy.asarray(depth1.sum(axis=-1)).dtype
    assert numpy.prod(array, axis=-1).dtype == numpy.asarray(depth1.prod(axis=-1)).dtype
    assert sum(awkward1.tolist(numpy.sum(array, axis=-1))) == sum(awkward1.tolist(depth1.sum(axis=-1)))
    assert prod(awkward1.tolist(numpy.prod(array, axis=-1))) == prod(awkward1.tolist(depth1.prod(axis=-1)))

    array = numpy.array([[0, 1, 2], [3, 4, 5]], dtype=numpy.uint64)
    content2 = awkward1.layout.NumpyArray(array.reshape(-1))
    offsets3 = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6], dtype=numpy.int64))
    depth1 = awkward1.layout.ListOffsetArray64(offsets3, content2)
    assert numpy.sum(array, axis=-1).dtype == numpy.asarray(depth1.sum(axis=-1)).dtype
    assert numpy.prod(array, axis=-1).dtype == numpy.asarray(depth1.prod(axis=-1)).dtype
    assert sum(awkward1.tolist(numpy.sum(array, axis=-1))) == sum(awkward1.tolist(depth1.sum(axis=-1)))
    assert prod(awkward1.tolist(numpy.prod(array, axis=-1))) == prod(awkward1.tolist(depth1.prod(axis=-1)))

def test_any():
    content2 = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3, 0.0, 2.2, 0.0, 0.0, 0.0, 0.0, 0.0]))
    offsets3 = awkward1.layout.Index64(numpy.array([0, 3, 6, 10], dtype=numpy.int64))
    depth1 = awkward1.layout.ListOffsetArray64(offsets3, content2)
    assert awkward1.tolist(depth1) == [
        [1.1, 2.2, 3.3],
        [0.0, 2.2, 0.0],
        [0.0, 0.0, 0.0, 0.0]]

    assert awkward1.tolist(depth1.any(-1)) == [
        True,
        True,
        False]
    assert awkward1.tolist(depth1.any(1)) == [
        True,
        True,
        False]

    assert awkward1.tolist(depth1.any(-2)) == [
        True,
        True,
        True,
        False]
    assert awkward1.tolist(depth1.any(0)) == [
        True,
        True,
        True,
        False]

def test_all():
    content2 = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3, 0.0, 2.2, 0.0, 0.0, 2.2, 0.0, 4.4]))
    offsets3 = awkward1.layout.Index64(numpy.array([0, 3, 6, 10], dtype=numpy.int64))
    depth1 = awkward1.layout.ListOffsetArray64(offsets3, content2)
    assert awkward1.tolist(depth1) == [
        [1.1, 2.2, 3.3],
        [0.0, 2.2, 0.0],
        [0.0, 2.2, 0.0, 4.4]]

    assert awkward1.tolist(depth1.all(-1)) == [
        True,
        False,
        False]
    assert awkward1.tolist(depth1.all(1)) == [
        True,
        False,
        False]

    assert awkward1.tolist(depth1.all(-2)) == [
        False,
        True,
        False,
        True]
    assert awkward1.tolist(depth1.all(0)) == [
        False,
        True,
        False,
        True]

def test_count():
    content2 = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3, 0.0, 2.2, 0.0, 0.0, 2.2, 0.0, 4.4]))
    offsets3 = awkward1.layout.Index64(numpy.array([0, 3, 6, 10], dtype=numpy.int64))
    depth1 = awkward1.layout.ListOffsetArray64(offsets3, content2)
    assert awkward1.tolist(depth1) == [
        [1.1, 2.2, 3.3],
        [0.0, 2.2, 0.0],
        [0.0, 2.2, 0.0, 4.4]]

    assert awkward1.tolist(depth1.count(-1)) == [
        3,
        3,
        4]
    assert awkward1.tolist(depth1.count(1)) == [
        3,
        3,
        4]

    assert awkward1.tolist(depth1.count(-2)) == [
        3,
        3,
        3,
        1]
    assert awkward1.tolist(depth1.count(0)) == [
        3,
        3,
        3,
        1]

def test_count_nonzero():
    content2 = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3, 0.0, 2.2, 0.0, 0.0, 2.2, 0.0, 4.4]))
    offsets3 = awkward1.layout.Index64(numpy.array([0, 3, 6, 10], dtype=numpy.int64))
    depth1 = awkward1.layout.ListOffsetArray64(offsets3, content2)
    assert awkward1.tolist(depth1) == [
        [1.1, 2.2, 3.3],
        [0.0, 2.2, 0.0],
        [0.0, 2.2, 0.0, 4.4]]

    assert awkward1.tolist(depth1.count_nonzero(-1)) == [
        3,
        1,
        2]
    assert awkward1.tolist(depth1.count_nonzero(1)) == [
        3,
        1,
        2]

    assert awkward1.tolist(depth1.count_nonzero(-2)) == [
        1,
        3,
        1,
        1]
    assert awkward1.tolist(depth1.count_nonzero(0)) == [
        1,
        3,
        1,
        1]

def test_count_min():
    content2 = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3, 0.0, 2.2, 0.0, 0.0, 2.2, 0.0, 4.4]))
    offsets3 = awkward1.layout.Index64(numpy.array([0, 3, 6, 10], dtype=numpy.int64))
    depth1 = awkward1.layout.ListOffsetArray64(offsets3, content2)
    assert awkward1.tolist(depth1) == [
        [1.1, 2.2, 3.3],
        [0.0, 2.2, 0.0],
        [0.0, 2.2, 0.0, 4.4]]

    assert awkward1.tolist(depth1.min(-1)) == [
        1.1,
        0.0,
        0.0]
    assert awkward1.tolist(depth1.min(1)) == [
        1.1,
        0.0,
        0.0]

    assert awkward1.tolist(depth1.min(-2)) == [
        0.0,
        2.2,
        0.0,
        4.4]
    assert awkward1.tolist(depth1.min(0)) == [
        0.0,
        2.2,
        0.0,
        4.4]

    content2 = awkward1.layout.NumpyArray(numpy.array([True, True, True, False, True, False, False, True, False, True]))
    offsets3 = awkward1.layout.Index64(numpy.array([0, 3, 6, 10], dtype=numpy.int64))
    depth1 = awkward1.layout.ListOffsetArray64(offsets3, content2)
    assert awkward1.tolist(depth1) == [
        [ True,  True,  True],
        [False,  True, False],
        [False,  True, False, True]]

    assert awkward1.tolist(depth1.min(-1)) == [
        True,
        False,
        False]
    assert awkward1.tolist(depth1.min(1)) == [
        True,
        False,
        False]

    assert awkward1.tolist(depth1.min(-2)) == [
        False,
        True,
        False,
        True]
    assert awkward1.tolist(depth1.min(0)) == [
        False,
        True,
        False,
        True]

def test_count_max():
    content2 = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3, 0.0, 2.2, 0.0, 0.0, 2.2, 0.0, 4.4]))
    offsets3 = awkward1.layout.Index64(numpy.array([0, 3, 6, 10], dtype=numpy.int64))
    depth1 = awkward1.layout.ListOffsetArray64(offsets3, content2)
    assert awkward1.tolist(depth1) == [
        [1.1, 2.2, 3.3],
        [0.0, 2.2, 0.0],
        [0.0, 2.2, 0.0, 4.4]]

    assert awkward1.tolist(depth1.max(-1)) == [
        3.3,
        2.2,
        4.4]
    assert awkward1.tolist(depth1.max(1)) == [
        3.3,
        2.2,
        4.4]

    assert awkward1.tolist(depth1.max(-2)) == [
        1.1,
        2.2,
        3.3,
        4.4]
    assert awkward1.tolist(depth1.max(0)) == [
        1.1,
        2.2,
        3.3,
        4.4]

    content2 = awkward1.layout.NumpyArray(numpy.array([False, True, True, False, True, False, False, False, False, False]))
    offsets3 = awkward1.layout.Index64(numpy.array([0, 3, 6, 10], dtype=numpy.int64))
    depth1 = awkward1.layout.ListOffsetArray64(offsets3, content2)
    assert awkward1.tolist(depth1) == [
        [False,  True,  True],
        [False,  True, False],
        [False, False, False, False]]

    assert awkward1.tolist(depth1.max(-1)) == [
        True,
        True,
        False]
    assert awkward1.tolist(depth1.max(1)) == [
        True,
        True,
        False]

    assert awkward1.tolist(depth1.max(-2)) == [
        False,
        True,
        True,
        False]
    assert awkward1.tolist(depth1.max(0)) == [
        False,
        True,
        True,
        False]

def test_mask():
    content = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]))
    offsets = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6, 6, 6, 9], dtype=numpy.int64))
    array = awkward1.layout.ListOffsetArray64(offsets, content)

    assert awkward1.tolist(array.min(axis=-1, mask=False)) == [1.1, numpy.inf, 4.4, 6.6, numpy.inf, numpy.inf, 7.7]
    assert awkward1.tolist(array.min(axis=-1, mask=True)) == [1.1, None, 4.4, 6.6, None, None, 7.7]

def test_keepdims():
    nparray = numpy.array(primes[:2*3*5], dtype=numpy.int64).reshape(2, 3, 5)
    content1 = awkward1.layout.NumpyArray(numpy.array(primes[:2*3*5], dtype=numpy.int64))
    offsets1 = awkward1.layout.Index64(numpy.array([0, 5, 10, 15, 20, 25, 30], dtype=numpy.int64))
    offsets2 = awkward1.layout.Index64(numpy.array([0, 3, 6], dtype=numpy.int64))
    depth2 = awkward1.layout.ListOffsetArray64(offsets2, awkward1.layout.ListOffsetArray64(offsets1, content1))
    assert awkward1.tolist(depth2) == [
        [[  2,   3,   5,   7,  11],
         [ 13,  17,  19,  23,  29],
         [ 31,  37,  41,  43,  47]],
        [[ 53,  59,  61,  67,  71],
         [ 73,  79,  83,  89,  97],
         [101, 103, 107, 109, 113]]]

    assert awkward1.tolist(depth2.prod(axis=-1, keepdims=False)) == awkward1.tolist(nparray.prod(axis=-1, keepdims=False))
    assert awkward1.tolist(depth2.prod(axis=-2, keepdims=False)) == awkward1.tolist(nparray.prod(axis=-2, keepdims=False))
    assert awkward1.tolist(depth2.prod(axis=-3, keepdims=False)) == awkward1.tolist(nparray.prod(axis=-3, keepdims=False))

    assert awkward1.tolist(depth2.prod(axis=-1, keepdims=True)) == awkward1.tolist(nparray.prod(axis=-1, keepdims=True))
    assert awkward1.tolist(depth2.prod(axis=-2, keepdims=True)) == awkward1.tolist(nparray.prod(axis=-2, keepdims=True))
    assert awkward1.tolist(depth2.prod(axis=-3, keepdims=True)) == awkward1.tolist(nparray.prod(axis=-3, keepdims=True))

def test_highlevel():
    array = awkward1.Array([
        [[ 2,  3, 5],
         [         ],
         [ 7, 11   ],
         [13       ]],
        [],
        [[17, 19   ],
         [23       ]]])

    assert awkward1.count(array) == 9
    assert awkward1.tolist(awkward1.count(array, axis=-1)) == [
        [3, 0, 2, 1],
        [],
        [2, 1]]
    assert awkward1.tolist(awkward1.count(array, axis=2)) == [
        [3, 0, 2, 1],
        [],
        [2, 1]]
    assert awkward1.tolist(awkward1.count(array, axis=-1, keepdims=True)) == [
        [[3], [0], [2], [1]],
        [],
        [[2], [1]]]
    assert awkward1.tolist(awkward1.count(array, axis=-2)) == [
        [3, 2, 1],
        [],
        [2, 1]]
    assert awkward1.tolist(awkward1.count(array, axis=1)) == [
        [3, 2, 1],
        [],
        [2, 1]]
    assert awkward1.tolist(awkward1.count(array, axis=-2, keepdims=True)) == [
        [[3, 2, 1]],
        [[]],
        [[2, 1]]]

    assert awkward1.count_nonzero(array) == 9
    assert awkward1.tolist(awkward1.count_nonzero(array, axis=-1)) == [
        [3, 0, 2, 1],
        [],
        [2, 1]]
    assert awkward1.tolist(awkward1.count_nonzero(array, axis=-2)) == [
        [3, 2, 1],
        [],
        [2, 1]]

    assert awkward1.sum(array) == 2 + 3 + 5 + 7 + 11 + 13 + 17 + 19 + 23
    assert awkward1.tolist(awkward1.sum(array, axis=-1)) == [
        [2 + 3 + 5, 0, 7 + 11, 13],
        [],
        [17 + 19, 23]]
    assert awkward1.tolist(awkward1.sum(array, axis=-2)) == [
        [2 + 7 + 13, 3 + 11, 5],
        [],
        [17 + 23, 19]]

    assert awkward1.prod(array) == 2*3*5*7*11*13*17*19*23
    assert awkward1.tolist(awkward1.prod(array, axis=-1)) == [
        [2*3*5, 1, 7*11, 13],
        [],
        [17*19, 23]]
    assert awkward1.tolist(awkward1.prod(array, axis=-2)) == [
        [2*7*13, 3*11, 5],
        [],
        [17*23, 19]]

    assert awkward1.min(array) == 2
    assert awkward1.tolist(awkward1.min(array, axis=-1)) == [
        [2, None, 7, 13],
        [],
        [17, 23]]
    assert awkward1.tolist(awkward1.min(array, axis=-2)) == [
        [2, 3, 5],
        [],
        [17, 19]]

    assert awkward1.max(array) == 23
    assert awkward1.tolist(awkward1.max(array, axis=-1)) == [
        [5, None, 11, 13],
        [],
        [19, 23]]
    assert awkward1.tolist(awkward1.max(array, axis=-2)) == [
        [13, 11, 5],
        [],
        [23, 19]]

    array = awkward1.Array([
        [[ True, False, True],
         [                  ],
         [False, False      ],
         [ True             ]],
        [],
        [[False,  True      ],
         [ True             ]]])

    assert awkward1.any(array) == True
    assert awkward1.tolist(awkward1.any(array, axis=-1)) == [
        [True, False, False, True],
        [],
        [True, True]]
    assert awkward1.tolist(awkward1.any(array, axis=-2)) == [
        [True, False, True],
        [],
        [True, True]]

    assert awkward1.all(array) == False
    assert awkward1.tolist(awkward1.all(array, axis=-1)) == [
        [False, True, False, True],
        [],
        [False, True]]
    assert awkward1.tolist(awkward1.all(array, axis=-2)) == [
        [False, False, True],
        [],
        [False, True]]

def test_nonreducers():
    x = awkward1.Array([[1, 2, 3, 4, 5], [1, 2, 3, 4, 5]])
    y = awkward1.Array([[1.1, 2.2, 2.9, 4.0, 5.1], [0.9, 2.1, 3.2, 4.1, 4.9]])

    assert awkward1.mean(y) == numpy.mean(awkward1.tonumpy(y))
    assert awkward1.var(y) == numpy.var(awkward1.tonumpy(y))
    assert awkward1.var(y, ddof=1) == numpy.var(awkward1.tonumpy(y), ddof=1)
    assert awkward1.std(y) == numpy.std(awkward1.tonumpy(y))
    assert awkward1.std(y, ddof=1) == numpy.std(awkward1.tonumpy(y), ddof=1)

    assert awkward1.moment(y, 1) == numpy.mean(awkward1.tonumpy(y))
    assert awkward1.moment(y - awkward1.mean(y), 2) == numpy.var(awkward1.tonumpy(y))
    assert awkward1.covar(y, y) == numpy.var(awkward1.tonumpy(y))
    assert awkward1.corr(y, y) == 1.0

    assert awkward1.corr(x, y) == pytest.approx(0.9968772535047296)
    fit = awkward1.linearfit(x, y)
    assert awkward1.tolist(fit) == pytest.approx({"intercept": 0.07999999999999773, "slope": 0.99, "intercept_error": 0.7416198487095663, "slope_error": 0.22360679774997896})

    assert awkward1.tolist(awkward1.mean(y, axis=-1)) == awkward1.tolist(numpy.mean(awkward1.tonumpy(y), axis=-1))
    assert awkward1.tolist(awkward1.var(y, axis=-1)) == awkward1.tolist(numpy.var(awkward1.tonumpy(y), axis=-1))
    assert awkward1.tolist(awkward1.var(y, axis=-1, ddof=1)) == awkward1.tolist(numpy.var(awkward1.tonumpy(y), axis=-1, ddof=1))
    assert awkward1.tolist(awkward1.std(y, axis=-1)) == awkward1.tolist(numpy.std(awkward1.tonumpy(y), axis=-1))
    assert awkward1.tolist(awkward1.std(y, axis=-1, ddof=1)) == awkward1.tolist(numpy.std(awkward1.tonumpy(y), axis=-1, ddof=1))

    assert awkward1.tolist(awkward1.moment(y, 1, axis=-1)) == awkward1.tolist(numpy.mean(awkward1.tonumpy(y), axis=-1))
    assert awkward1.tolist(awkward1.moment(y - awkward1.mean(y, axis=-1), 2, axis=-1)) == awkward1.tolist(numpy.var(awkward1.tonumpy(y), axis=-1))
    assert awkward1.tolist(awkward1.covar(y, y, axis=-1)) == awkward1.tolist(numpy.var(awkward1.tonumpy(y), axis=-1))
    assert awkward1.tolist(awkward1.corr(y, y, axis=-1)) == [1.0, 1.0]

    assert awkward1.tolist(awkward1.corr(x, y, axis=-1)) == pytest.approx([0.9975103695813371, 0.9964193240901015])
    fit = awkward1.linearfit(x, y, axis=-1)
    assert awkward1.tolist(fit[0]) == pytest.approx({"intercept": 0.11999999999999772, "slope": 0.9800000000000005, "intercept_error": 1.0488088481701516, "slope_error": 0.31622776601683794})
    assert awkward1.tolist(fit[1]) == pytest.approx({"intercept": 0.04000000000000228, "slope": 0.9999999999999994, "intercept_error": 1.0488088481701516, "slope_error": 0.31622776601683794})

def test_softmax():
    array = awkward1.Array([[numpy.log(2), numpy.log(2), numpy.log(4)], [], [numpy.log(5), numpy.log(5)]])
    assert awkward1.tolist(awkward1.softmax(array, axis=-1)) == [pytest.approx([0.25, 0.25, 0.5]), [], pytest.approx([0.5, 0.5])]



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import json

import numpy
import numba

import awkward1.layout
import awkward1._util
import awkward1._numba.arrayview

@numba.extending.typeof_impl.register(awkward1.layout.NumpyArray)
def typeof(obj, c):
    return NumpyArrayType(numba.typeof(numpy.asarray(obj)), numba.typeof(obj.identities), obj.parameters)

@numba.extending.typeof_impl.register(awkward1.layout.RegularArray)
def typeof(obj, c):
    return RegularArrayType(numba.typeof(obj.content), obj.size, numba.typeof(obj.identities), obj.parameters)

@numba.extending.typeof_impl.register(awkward1.layout.ListArray32)
@numba.extending.typeof_impl.register(awkward1.layout.ListArrayU32)
@numba.extending.typeof_impl.register(awkward1.layout.ListArray64)
@numba.extending.typeof_impl.register(awkward1.layout.ListOffsetArray32)
@numba.extending.typeof_impl.register(awkward1.layout.ListOffsetArrayU32)
@numba.extending.typeof_impl.register(awkward1.layout.ListOffsetArray64)
def typeof(obj, c):
    return ListArrayType(numba.typeof(numpy.asarray(obj.starts)), numba.typeof(obj.content), numba.typeof(obj.identities), obj.parameters)

@numba.extending.typeof_impl.register(awkward1.layout.IndexedArray32)
@numba.extending.typeof_impl.register(awkward1.layout.IndexedArrayU32)
@numba.extending.typeof_impl.register(awkward1.layout.IndexedArray64)
def typeof(obj, c):
    return IndexedArrayType(numba.typeof(numpy.asarray(obj.index)), numba.typeof(obj.content), numba.typeof(obj.identities), obj.parameters)

@numba.extending.typeof_impl.register(awkward1.layout.IndexedOptionArray32)
@numba.extending.typeof_impl.register(awkward1.layout.IndexedOptionArray64)
def typeof(obj, c):
    return IndexedOptionArrayType(numba.typeof(numpy.asarray(obj.index)), numba.typeof(obj.content), numba.typeof(obj.identities), obj.parameters)

@numba.extending.typeof_impl.register(awkward1.layout.RecordArray)
def typeof(obj, c):
    return RecordArrayType(tuple(numba.typeof(x) for x in obj.contents), obj.recordlookup, numba.typeof(obj.identities), obj.parameters)

@numba.extending.typeof_impl.register(awkward1.layout.Record)
def typeof(obj, c):
    return RecordType(numba.typeof(obj.array))

@numba.extending.typeof_impl.register(awkward1.layout.UnionArray8_32)
@numba.extending.typeof_impl.register(awkward1.layout.UnionArray8_U32)
@numba.extending.typeof_impl.register(awkward1.layout.UnionArray8_64)
def typeof(obj, c):
    return UnionArrayType(numba.typeof(numpy.asarray(obj.tags)), numba.typeof(numpy.asarray(obj.index)), tuple(numba.typeof(x) for x in obj.contents), numba.typeof(obj.identities), obj.parameters)

class ContentType(numba.types.Type):
    @classmethod
    def tolookup_identities(cls, layout, positions, sharedptrs, arrays):
        if layout.identities is None:
            positions.append(-1)
            sharedptrs.append(None)
        else:
            arrays.append(numpy.asarray(layout.identities))
            positions.append(arrays[-1])
            sharedptrs.append(None)

    def IndexOf(self, arraytype):
        if arraytype.dtype.bitwidth == 8 and arraytype.dtype.signed:
            return awkward1.layout.Index8
        elif arraytype.dtype.bitwidth == 8:
            return awkward1.layout.IndexU8
        elif arraytype.dtype.bitwidth == 32 and arraytype.dtype.signed:
            return awkward1.layout.Index32
        elif arraytype.dtype.bitwidth == 32:
            return awkward1.layout.IndexU32
        elif arraytype.dtype.bitwidth == 64 and arraytype.dtype.signed:
            return awkward1.layout.Index64
        else:
            raise AssertionError("no Index* type for array: {0}".format(arraytype))

    def getitem_at_check(self, viewtype):
        typer = awkward1._util.numba_array_typer(viewtype.type, viewtype.behavior)
        if typer is None:
            return self.getitem_at(viewtype)
        else:
            return typer(viewtype)

    def getitem_range(self, viewtype):
        return awkward1._numba.arrayview.wrap(self, viewtype, None)

    def getitem_field(self, viewtype, key):
        if self.hasfield(key):
            return awkward1._numba.arrayview.wrap(self, viewtype, viewtype.fields + (key,))
        else:
            raise TypeError("array does not have a field with key {0}".format(repr(key)))

    def lower_getitem_at_check(self, context, builder, rettype, viewtype, viewval, viewproxy, attype, atval, wrapneg, checkbounds):
        lower = awkward1._util.numba_array_lower(viewtype.type, viewtype.behavior)
        if lower is not None:
            atval = regularize_atval(context, builder, viewproxy, attype, atval, wrapneg, checkbounds)
            return lower(context, builder, rettype, viewtype, viewval, viewproxy, attype, atval)
        else:
            return self.lower_getitem_at(context, builder, rettype, viewtype, viewval, viewproxy, attype, atval, wrapneg, checkbounds)

    def lower_getitem_range(self, context, builder, rettype, viewtype, viewval, viewproxy, start, stop, wrapneg):
        length = builder.sub(viewproxy.stop, viewproxy.start)

        regular_start = numba.cgutils.alloca_once_value(builder, start)
        regular_stop = numba.cgutils.alloca_once_value(builder, stop)

        if wrapneg:
            with builder.if_then(builder.icmp_signed("<", start, context.get_constant(numba.intp, 0))):
                builder.store(builder.add(start, length), regular_start)
            with builder.if_then(builder.icmp_signed("<", stop, context.get_constant(numba.intp, 0))):
                builder.store(builder.add(stop, length), regular_stop)

        with builder.if_then(builder.icmp_signed("<", builder.load(regular_start), context.get_constant(numba.intp, 0))):
            builder.store(context.get_constant(numba.intp, 0), regular_start)
        with builder.if_then(builder.icmp_signed(">", builder.load(regular_start), length)):
            builder.store(length, regular_start)

        with builder.if_then(builder.icmp_signed("<", builder.load(regular_stop), builder.load(regular_start))):
            builder.store(builder.load(regular_start), regular_stop)
        with builder.if_then(builder.icmp_signed(">", builder.load(regular_stop), length)):
            builder.store(length, regular_stop)

        proxyout = context.make_helper(builder, rettype)
        proxyout.pos        = viewproxy.pos
        proxyout.start      = builder.add(viewproxy.start, builder.load(regular_start))
        proxyout.stop       = builder.add(viewproxy.start, builder.load(regular_stop))
        proxyout.arrayptrs  = viewproxy.arrayptrs
        proxyout.sharedptrs = viewproxy.sharedptrs
        proxyout.pylookup   = viewproxy.pylookup
        return proxyout._getvalue()

    def lower_getitem_field(self, context, builder, viewtype, viewval, key):
        return viewval

def posat(context, builder, pos, offset):
    return builder.add(pos, context.get_constant(numba.intp, offset))

def getat(context, builder, baseptr, offset, rettype=None):
    ptrtype = None
    if rettype is not None:
        ptrtype = context.get_value_type(numba.types.CPointer(rettype))
        bitwidth = rettype.bitwidth
    else:
        bitwidth = numba.intp.bitwidth
    byteoffset = builder.mul(offset, context.get_constant(numba.intp, bitwidth // 8))
    return builder.load(numba.cgutils.pointer_add(builder, baseptr, byteoffset, ptrtype))

def regularize_atval(context, builder, viewproxy, attype, atval, wrapneg, checkbounds):
    atval = awkward1._numba.castint(context, builder, attype, numba.intp, atval)

    if not attype.signed:
        wrapneg = False

    if wrapneg or checkbounds:
        length = builder.sub(viewproxy.stop, viewproxy.start)

        if wrapneg:
            regular_atval = numba.cgutils.alloca_once_value(builder, atval)
            with builder.if_then(builder.icmp_signed("<", atval, context.get_constant(numba.intp, 0))):
                builder.store(builder.add(atval, length), regular_atval)
            atval = builder.load(regular_atval)

        if checkbounds:
            with builder.if_then(builder.or_(builder.icmp_signed("<", atval, context.get_constant(numba.intp, 0)),
                                             builder.icmp_signed(">=", atval, length))):
                context.call_conv.return_user_exc(builder, ValueError, ("slice index out of bounds",))

    return awkward1._numba.castint(context, builder, atval.type, numba.intp, atval)

class NumpyArrayType(ContentType):
    IDENTITIES = 0
    ARRAY = 1

    @classmethod
    def tolookup(cls, layout, positions, sharedptrs, arrays):
        array = numpy.asarray(layout)
        assert len(array.shape) == 1
        pos = len(positions)
        cls.tolookup_identities(layout, positions, sharedptrs, arrays)
        sharedptrs[-1] = layout._persistent_shared_ptr
        positions.append(array)
        sharedptrs.append(None)
        arrays.append(array)
        return pos

    def __init__(self, arraytype, identitiestype, parameters):
        super(NumpyArrayType, self).__init__(name="awkward1.NumpyArrayType({0}, {1}, {2})".format(arraytype.name, identitiestype.name, json.dumps(parameters)))
        self.arraytype = arraytype
        self.identitiestype = identitiestype
        self.parameters = parameters

    def tolayout(self, lookup, pos, fields):
        assert fields == ()
        return awkward1.layout.NumpyArray(lookup.arrays[lookup.positions[pos + self.ARRAY]], parameters=self.parameters)

    def hasfield(self, key):
        return False

    def getitem_at(self, viewtype):
        return self.arraytype.dtype

    def lower_getitem_at(self, context, builder, rettype, viewtype, viewval, viewproxy, attype, atval, wrapneg, checkbounds):
        whichpos = posat(context, builder, viewproxy.pos, self.ARRAY)
        arrayptr = getat(context, builder, viewproxy.arrayptrs, whichpos)
        atval = regularize_atval(context, builder, viewproxy, attype, atval, wrapneg, checkbounds)
        arraypos = builder.add(viewproxy.start, atval)
        return getat(context, builder, arrayptr, arraypos, rettype)

class RegularArrayType(ContentType):
    IDENTITIES = 0
    CONTENT = 1

    @classmethod
    def tolookup(cls, layout, positions, sharedptrs, arrays):
        pos = len(positions)
        cls.tolookup_identities(layout, positions, sharedptrs, arrays)
        sharedptrs[-1] = layout._persistent_shared_ptr
        positions.append(None)
        sharedptrs.append(None)
        positions[pos + cls.CONTENT] = awkward1._numba.arrayview.tolookup(layout.content, positions, sharedptrs, arrays)
        return pos

    def __init__(self, contenttype, size, identitiestype, parameters):
        super(RegularArrayType, self).__init__(name="awkward1.RegularArrayType({0}, {1}, {2}, {3})".format(contenttype.name, size, identitiestype.name, json.dumps(parameters)))
        self.contenttype = contenttype
        self.size = size
        self.identitiestype = identitiestype
        self.parameters = parameters

    def tolayout(self, lookup, pos, fields):
        content = self.contenttype.tolayout(lookup, lookup.positions[pos + self.CONTENT], fields)
        return awkward1.layout.RegularArray(content, self.size, parameters=self.parameters)

    def hasfield(self, key):
        return self.contenttype.hasfield(key)

    def getitem_at(self, viewtype):
        return awkward1._numba.arrayview.wrap(self.contenttype, viewtype, None)

    def lower_getitem_at(self, context, builder, rettype, viewtype, viewval, viewproxy, attype, atval, wrapneg, checkbounds):
        whichpos = posat(context, builder, viewproxy.pos, self.CONTENT)
        nextpos = getat(context, builder, viewproxy.arrayptrs, whichpos)

        atval = regularize_atval(context, builder, viewproxy, attype, atval, wrapneg, checkbounds)

        size = context.get_constant(numba.intp, self.size)
        start = builder.mul(builder.add(viewproxy.start, atval), size)
        stop  = builder.add(start, size)

        proxyout = context.make_helper(builder, rettype)
        proxyout.pos        = nextpos
        proxyout.start      = start
        proxyout.stop       = stop
        proxyout.arrayptrs  = viewproxy.arrayptrs
        proxyout.sharedptrs = viewproxy.sharedptrs
        proxyout.pylookup   = viewproxy.pylookup
        return proxyout._getvalue()

class ListArrayType(ContentType):
    IDENTITIES = 0
    STARTS = 1
    STOPS = 2
    CONTENT = 3

    @classmethod
    def tolookup(cls, layout, positions, sharedptrs, arrays):
        if isinstance(layout, (awkward1.layout.ListArray32, awkward1.layout.ListArrayU32, awkward1.layout.ListArray64)):
            starts = numpy.asarray(layout.starts)
            stops = numpy.asarray(layout.stops)
        elif isinstance(layout, (awkward1.layout.ListOffsetArray32, awkward1.layout.ListOffsetArrayU32, awkward1.layout.ListOffsetArray64)):
            offsets = numpy.asarray(layout.offsets)
            starts = offsets[:-1]
            stops = offsets[1:]

        pos = len(positions)
        cls.tolookup_identities(layout, positions, sharedptrs, arrays)
        sharedptrs[-1] = layout._persistent_shared_ptr
        positions.append(starts)
        sharedptrs.append(None)
        arrays.append(starts)
        positions.append(stops)
        sharedptrs.append(None)
        arrays.append(stops)
        positions.append(None)
        sharedptrs.append(None)
        positions[pos + cls.CONTENT] = awkward1._numba.arrayview.tolookup(layout.content, positions, sharedptrs, arrays)
        return pos

    def __init__(self, indextype, contenttype, identitiestype, parameters):
        super(ListArrayType, self).__init__(name="awkward1.ListArrayType({0}, {1}, {2}, {3})".format(indextype.name, contenttype.name, identitiestype.name, json.dumps(parameters)))
        self.indextype = indextype
        self.contenttype = contenttype
        self.identitiestype = identitiestype
        self.parameters = parameters

    def ListArrayOf(self):
        if self.indextype.dtype.bitwidth == 32 and self.indextype.dtype.signed:
            return awkward1.layout.ListArray32
        elif self.indextype.dtype.bitwidth == 32:
            return awkward1.layout.ListArrayU32
        elif self.indextype.dtype.bitwidth == 64 and self.indextype.dtype.signed:
            return awkward1.layout.ListArray64
        else:
            raise AssertionError("no ListArray* type for array: {0}".format(indextype))

    def tolayout(self, lookup, pos, fields):
        starts = self.IndexOf(self.indextype)(lookup.arrays[lookup.positions[pos + self.STARTS]])
        stops = self.IndexOf(self.indextype)(lookup.arrays[lookup.positions[pos + self.STOPS]])
        content = self.contenttype.tolayout(lookup, lookup.positions[pos + self.CONTENT], fields)
        return self.ListArrayOf()(starts, stops, content, parameters=self.parameters)

    def hasfield(self, key):
        return self.contenttype.hasfield(key)

    def getitem_at(self, viewtype):
        return awkward1._numba.arrayview.wrap(self.contenttype, viewtype, None)

    def lower_getitem_at(self, context, builder, rettype, viewtype, viewval, viewproxy, attype, atval, wrapneg, checkbounds):
        whichpos = posat(context, builder, viewproxy.pos, self.CONTENT)
        nextpos = getat(context, builder, viewproxy.arrayptrs, whichpos)

        atval = regularize_atval(context, builder, viewproxy, attype, atval, wrapneg, checkbounds)

        startspos = posat(context, builder, viewproxy.pos, self.STARTS)
        startsptr = getat(context, builder, viewproxy.arrayptrs, startspos)
        startsarraypos = builder.add(viewproxy.start, atval)
        start = getat(context, builder, startsptr, startsarraypos, self.indextype.dtype)

        stopspos = posat(context, builder, viewproxy.pos, self.STOPS)
        stopsptr = getat(context, builder, viewproxy.arrayptrs, stopspos)
        stopsarraypos = builder.add(viewproxy.start, atval)
        stop = getat(context, builder, stopsptr, stopsarraypos, self.indextype.dtype)

        proxyout = context.make_helper(builder, rettype)
        proxyout.pos        = nextpos
        proxyout.start      = awkward1._numba.castint(context, builder, self.indextype.dtype, numba.intp, start)
        proxyout.stop       = awkward1._numba.castint(context, builder, self.indextype.dtype, numba.intp, stop)
        proxyout.arrayptrs  = viewproxy.arrayptrs
        proxyout.sharedptrs = viewproxy.sharedptrs
        proxyout.pylookup   = viewproxy.pylookup
        return proxyout._getvalue()

class IndexedArrayType(ContentType):
    IDENTITIES = 0
    INDEX = 1
    CONTENT = 2

    @classmethod
    def tolookup(cls, layout, positions, sharedptrs, arrays):
        pos = len(positions)
        cls.tolookup_identities(layout, positions, sharedptrs, arrays)
        sharedptrs[-1] = layout._persistent_shared_ptr
        arrays.append(numpy.asarray(layout.index))
        positions.append(arrays[-1])
        sharedptrs.append(None)
        positions.append(None)
        sharedptrs.append(None)
        positions[pos + cls.CONTENT] = awkward1._numba.arrayview.tolookup(layout.content, positions, sharedptrs, arrays)
        return pos

    def __init__(self, indextype, contenttype, identitiestype, parameters):
        super(IndexedArrayType, self).__init__(name="awkward1.IndexedArrayType({0}, {1}, {2}, {3})".format(indextype.name, contenttype.name, identitiestype.name, json.dumps(parameters)))
        self.indextype = indextype
        self.contenttype = contenttype
        self.identitiestype = identitiestype
        self.parameters = parameters

    def IndexedArrayOf(self):
        if self.indextype.dtype.bitwidth == 32 and self.indextype.dtype.signed:
            return awkward1.layout.IndexedArray32
        elif self.indextype.dtype.bitwidth == 32:
            return awkward1.layout.IndexedArrayU32
        elif self.indextype.dtype.bitwidth == 64 and self.indextype.dtype.signed:
            return awkward1.layout.IndexedArray64
        else:
            raise AssertionError("no IndexedArray* type for array: {0}".format(self.indextype))

    def tolayout(self, lookup, pos, fields):
        index = self.IndexOf(self.indextype)(lookup.arrays[lookup.positions[pos + self.INDEX]])
        content = self.contenttype.tolayout(lookup, lookup.positions[pos + self.CONTENT], fields)
        return self.IndexedArrayOf()(index, content, parameters=self.parameters)

    def hasfield(self, key):
        return self.contenttype.hasfield(key)

    def getitem_at(self, viewtype):
        return self.contenttype.getitem_at_check(viewtype)

    def lower_getitem_at(self, context, builder, rettype, viewtype, viewval, viewproxy, attype, atval, wrapneg, checkbounds):
        whichpos = posat(context, builder, viewproxy.pos, self.CONTENT)
        nextpos = getat(context, builder, viewproxy.arrayptrs, whichpos)

        atval = regularize_atval(context, builder, viewproxy, attype, atval, wrapneg, checkbounds)

        indexpos = posat(context, builder, viewproxy.pos, self.INDEX)
        indexptr = getat(context, builder, viewproxy.arrayptrs, indexpos)
        indexarraypos = builder.add(viewproxy.start, atval)
        nextat = getat(context, builder, indexptr, indexarraypos, self.indextype.dtype)

        nextviewtype = awkward1._numba.arrayview.wrap(self.contenttype, viewtype, None)
        proxynext = context.make_helper(builder, nextviewtype)
        proxynext.pos        = nextpos
        proxynext.start      = viewproxy.start
        proxynext.stop       = builder.add(awkward1._numba.castint(context, builder, self.indextype.dtype, numba.intp, nextat), builder.add(viewproxy.start, context.get_constant(numba.intp, 1)))
        proxynext.arrayptrs  = viewproxy.arrayptrs
        proxynext.sharedptrs = viewproxy.sharedptrs
        proxynext.pylookup   = viewproxy.pylookup

        return self.contenttype.lower_getitem_at_check(context, builder, rettype, nextviewtype, proxynext._getvalue(), proxynext, numba.intp, nextat, False, False)

class IndexedOptionArrayType(ContentType):
    IDENTITIES = 0
    INDEX = 1
    CONTENT = 2

    @classmethod
    def tolookup(cls, layout, positions, sharedptrs, arrays):
        pos = len(positions)
        cls.tolookup_identities(layout, positions, sharedptrs, arrays)
        sharedptrs[-1] = layout._persistent_shared_ptr
        arrays.append(numpy.asarray(layout.index))
        positions.append(arrays[-1])
        sharedptrs.append(None)
        positions.append(None)
        sharedptrs.append(None)
        positions[pos + cls.CONTENT] = awkward1._numba.arrayview.tolookup(layout.content, positions, sharedptrs, arrays)
        return pos

    def __init__(self, indextype, contenttype, identitiestype, parameters):
        super(IndexedOptionArrayType, self).__init__(name="awkward1.IndexedOptionArrayType({0}, {1}, {2}, {3})".format(indextype.name, contenttype.name, identitiestype.name, json.dumps(parameters)))
        self.indextype = indextype
        self.contenttype = contenttype
        self.identitiestype = identitiestype
        self.parameters = parameters

    def IndexedOptionArrayOf(self):
        if self.indextype.dtype.bitwidth == 32 and self.indextype.dtype.signed:
            return awkward1.layout.IndexedOptionArray32
        elif self.indextype.dtype.bitwidth == 64 and self.indextype.dtype.signed:
            return awkward1.layout.IndexedOptionArray64
        else:
            raise AssertionError("no IndexedOptionArray* type for array: {0}".format(self.indextype))

    def tolayout(self, lookup, pos, fields):
        index = self.IndexOf(self.indextype)(lookup.arrays[lookup.positions[pos + self.INDEX]])
        content = self.contenttype.tolayout(lookup, lookup.positions[pos + self.CONTENT], fields)
        return self.IndexedOptionArrayOf()(index, content, parameters=self.parameters)

    def hasfield(self, key):
        return self.contenttype.hasfield(key)

    def getitem_at(self, viewtype):
        return numba.types.optional(self.contenttype.getitem_at_check(viewtype))

    def lower_getitem_at(self, context, builder, rettype, viewtype, viewval, viewproxy, attype, atval, wrapneg, checkbounds):
        whichpos = posat(context, builder, viewproxy.pos, self.CONTENT)
        nextpos = getat(context, builder, viewproxy.arrayptrs, whichpos)

        atval = regularize_atval(context, builder, viewproxy, attype, atval, wrapneg, checkbounds)

        indexpos = posat(context, builder, viewproxy.pos, self.INDEX)
        indexptr = getat(context, builder, viewproxy.arrayptrs, indexpos)
        indexarraypos = builder.add(viewproxy.start, atval)
        nextat = getat(context, builder, indexptr, indexarraypos, self.indextype.dtype)

        output = context.make_helper(builder, rettype)

        with builder.if_else(builder.icmp_signed("<", nextat, context.get_constant(self.indextype.dtype, 0))) as (isnone, isvalid):
            with isnone:
                output.valid = numba.cgutils.false_bit
                output.data = numba.cgutils.get_null_value(output.data.type)

            with isvalid:
                nextviewtype = awkward1._numba.arrayview.wrap(self.contenttype, viewtype, None)
                proxynext = context.make_helper(builder, nextviewtype)
                proxynext.pos        = nextpos
                proxynext.start      = viewproxy.start
                proxynext.stop       = builder.add(awkward1._numba.castint(context, builder, self.indextype.dtype, numba.intp, nextat), builder.add(viewproxy.start, context.get_constant(numba.intp, 1)))
                proxynext.arrayptrs  = viewproxy.arrayptrs
                proxynext.sharedptrs = viewproxy.sharedptrs
                proxynext.pylookup   = viewproxy.pylookup

                outdata = self.contenttype.lower_getitem_at_check(context, builder, rettype.type, nextviewtype, proxynext._getvalue(), proxynext, numba.intp, nextat, False, False)

                output.valid = numba.cgutils.true_bit
                output.data = outdata

        return output._getvalue()

class RecordArrayType(ContentType):
    IDENTITIES = 0
    CONTENTS = 1

    @classmethod
    def tolookup(cls, layout, positions, sharedptrs, arrays):
        pos = len(positions)
        cls.tolookup_identities(layout, positions, sharedptrs, arrays)
        sharedptrs[-1] = layout._persistent_shared_ptr
        positions.extend([None] * layout.numfields)
        sharedptrs.extend([None] * layout.numfields)
        for i, content in enumerate(layout.contents):
            positions[pos + cls.CONTENTS + i] = awkward1._numba.arrayview.tolookup(content, positions, sharedptrs, arrays)
        return pos

    def __init__(self, contenttypes, recordlookup, identitiestype, parameters):
        super(RecordArrayType, self).__init__(name="awkward1.RecordArrayType(({0}{1}), ({2}), {3}, {4})".format(", ".join(x.name for x in contenttypes), "," if len(contenttypes) == 1 else "", "None" if recordlookup is None else repr(tuple(recordlookup)), identitiestype.name, json.dumps(parameters)))
        self.contenttypes = contenttypes
        self.recordlookup = recordlookup
        self.identitiestype = identitiestype
        self.parameters = parameters

    def fieldindex(self, key):
        out = -1
        if self.recordlookup is not None:
            for i, x in enumerate(self.recordlookup):
                if x == key:
                    out = i
                    break
        if out == -1:
            try:
                out = int(key)
            except ValueError:
                return None
            if not 0 <= out < len(self.contenttypes):
                return None
        return out

    def tolayout(self, lookup, pos, fields):
        if len(fields) > 0:
            index = self.fieldindex(fields[0])
            assert index is not None
            return self.contenttypes[index].tolayout(lookup, lookup.positions[pos + self.CONTENTS + index], fields[1:])
        else:
            contents = []
            for i, contenttype in enumerate(self.contenttypes):
                layout = contenttype.tolayout(lookup, lookup.positions[pos + self.CONTENTS + i], fields)
                contents.append(layout)
            if len(contents) == 0:
                return awkward1.layout.RecordArray(numpy.iinfo(numpy.int64).max, self.recordlookup is None, parameters=self.parameters)
            else:
                return awkward1.layout.RecordArray(contents, self.recordlookup, parameters=self.parameters)

    def hasfield(self, key):
        return self.fieldindex(key) is not None

    def getitem_at_check(self, viewtype):
        out = self.getitem_at(viewtype)
        if isinstance(out, awkward1._numba.arrayview.RecordViewType):
            typer = awkward1._util.numba_record_typer(out.arrayviewtype.type, out.arrayviewtype.behavior)
            if typer is not None:
                return typer(out)
        return out

    def getitem_at(self, viewtype):
        if len(viewtype.fields) == 0:
            return awkward1._numba.arrayview.RecordViewType(viewtype)
        else:
            key = viewtype.fields[0]
            index = self.fieldindex(key)
            if index is None:
                if self.recordlookup is None:
                    raise ValueError("no field {0} in tuples with {1} fields".format(repr(key), len(self.contenttypes)))
                else:
                    raise ValueError("no field {0} in records with fields: [{1}]".format(repr(key), ", ".join(repr(x) for x in self.recordlookup)))
            contenttype = self.contenttypes[index]
            subviewtype = awkward1._numba.arrayview.wrap(contenttype, viewtype, viewtype.fields[1:])
            return contenttype.getitem_at_check(subviewtype)

    def getitem_field(self, viewtype, key):
        index = self.fieldindex(key)
        if index is None:
            if self.recordlookup is None:
                raise ValueError("no field {0} in tuples with {1} fields".format(repr(key), len(self.contenttypes)))
            else:
                raise ValueError("no field {0} in records with fields: [{1}]".format(repr(key), ", ".join(repr(x) for x in self.recordlookup)))
        contenttype = self.contenttypes[index]
        subviewtype = awkward1._numba.arrayview.wrap(contenttype, viewtype, None)
        return contenttype.getitem_range(subviewtype)

    def getitem_field_record(self, recordviewtype, key):
        index = self.fieldindex(key)
        if index is None:
            if self.recordlookup is None:
                raise ValueError("no field {0} in tuple with {1} fields".format(repr(key), len(self.contenttypes)))
            else:
                raise ValueError("no field {0} in record with fields: [{1}]".format(repr(key), ", ".join(repr(x) for x in self.recordlookup)))
        contenttype = self.contenttypes[index]
        subviewtype = awkward1._numba.arrayview.wrap(contenttype, recordviewtype, None)
        return contenttype.getitem_at_check(subviewtype)

    def lower_getitem_at_check(self, context, builder, rettype, viewtype, viewval, viewproxy, attype, atval, wrapneg, checkbounds):
        out = self.lower_getitem_at(context, builder, rettype, viewtype, viewval, viewproxy, attype, atval, wrapneg, checkbounds)
        baretype = self.getitem_at(viewtype)
        if isinstance(baretype, awkward1._numba.arrayview.RecordViewType):
            lower = awkward1._util.numba_record_lower(baretype.arrayviewtype.type, baretype.arrayviewtype.behavior)
            if lower is not None:
                return lower(context, builder, rettype(baretype), (out,))
        return out

    def lower_getitem_at(self, context, builder, rettype, viewtype, viewval, viewproxy, attype, atval, wrapneg, checkbounds):
        atval = regularize_atval(context, builder, viewproxy, attype, atval, wrapneg, checkbounds)

        if len(viewtype.fields) == 0:
            proxyout = context.make_helper(builder, awkward1._numba.arrayview.RecordViewType(viewtype))
            proxyout.arrayview = viewval
            proxyout.at        = atval
            return proxyout._getvalue()

        else:
            index = self.fieldindex(viewtype.fields[0])
            contenttype = self.contenttypes[index]

            whichpos = posat(context, builder, viewproxy.pos, self.CONTENTS + index)
            nextpos = getat(context, builder, viewproxy.arrayptrs, whichpos)

            nextviewtype = awkward1._numba.arrayview.wrap(contenttype, viewtype, viewtype.fields[1:])
            proxynext = context.make_helper(builder, nextviewtype)
            proxynext.pos        = nextpos
            proxynext.start      = viewproxy.start
            proxynext.stop       = builder.add(atval, builder.add(viewproxy.start, context.get_constant(numba.intp, 1)))
            proxynext.arrayptrs  = viewproxy.arrayptrs
            proxynext.sharedptrs = viewproxy.sharedptrs
            proxynext.pylookup   = viewproxy.pylookup

            return contenttype.lower_getitem_at_check(context, builder, rettype, nextviewtype, proxynext._getvalue(), proxynext, numba.intp, atval, False, False)

    def lower_getitem_field(self, context, builder, viewtype, viewval, key):
        viewproxy = context.make_helper(builder, viewtype, viewval)

        index = self.fieldindex(key)
        contenttype = self.contenttypes[index]

        whichpos = posat(context, builder, viewproxy.pos, self.CONTENTS + index)
        nextpos = getat(context, builder, viewproxy.arrayptrs, whichpos)

        proxynext = context.make_helper(builder, contenttype.getitem_range(viewtype))
        proxynext.pos        = nextpos
        proxynext.start      = viewproxy.start
        proxynext.stop       = viewproxy.stop
        proxynext.arrayptrs  = viewproxy.arrayptrs
        proxynext.sharedptrs = viewproxy.sharedptrs
        proxynext.pylookup   = viewproxy.pylookup

        return proxynext._getvalue()

    def lower_getitem_field_record(self, context, builder, recordviewtype, recordviewval, key):
        arrayviewtype = recordviewtype.arrayviewtype
        recordviewproxy = context.make_helper(builder, recordviewtype, recordviewval)
        arrayviewval = recordviewproxy.arrayview
        arrayviewproxy = context.make_helper(builder, arrayviewtype, arrayviewval)

        index = self.fieldindex(key)
        contenttype = self.contenttypes[index]

        whichpos = posat(context, builder, arrayviewproxy.pos, self.CONTENTS + index)
        nextpos = getat(context, builder, arrayviewproxy.arrayptrs, whichpos)

        proxynext = context.make_helper(builder, contenttype.getitem_range(arrayviewtype))
        proxynext.pos        = nextpos
        proxynext.start      = arrayviewproxy.start
        proxynext.stop       = builder.add(recordviewproxy.at, builder.add(arrayviewproxy.start, context.get_constant(numba.intp, 1)))
        proxynext.arrayptrs  = arrayviewproxy.arrayptrs
        proxynext.sharedptrs = arrayviewproxy.sharedptrs
        proxynext.pylookup   = arrayviewproxy.pylookup

        nextviewtype = awkward1._numba.arrayview.wrap(contenttype, arrayviewtype, None)

        rettype = self.getitem_field_record(recordviewtype, key)

        return contenttype.lower_getitem_at_check(context, builder, rettype, nextviewtype, proxynext._getvalue(), proxynext, numba.intp, recordviewproxy.at, False, False)

class UnionArrayType(ContentType):
    IDENTITIES = 0
    TAGS = 1
    INDEX = 2
    CONTENTS = 3

    @classmethod
    def tolookup(cls, layout, positions, sharedptrs, arrays):
        pos = len(positions)
        cls.tolookup_identities(layout, positions, sharedptrs, arrays)
        sharedptrs[-1] = layout._persistent_shared_ptr
        arrays.append(numpy.asarray(layout.tags))
        positions.append(arrays[-1])
        sharedptrs.append(None)
        arrays.append(numpy.asarray(layout.index))
        positions.append(arrays[-1])
        sharedptrs.append(None)
        positions.extend([None] * layout.numcontents)
        sharedptrs.extend([None] * layout.numcontents)
        for i, content in enumerate(layout.contents):
            positions[pos + cls.CONTENTS + i] = awkward1._numba.arrayview.tolookup(content, positions, sharedptrs, arrays)
        return pos

    def __init__(self, tagstype, indextype, contenttypes, identitiestype, parameters):
        super(UnionArrayType, self).__init__(name="awkward1.UnionArrayType({0}, {1}, ({2}{3}), {4}, {5})".format(tagstype.name, indextype.name, ", ".join(x.name for x in contenttypes), "," if len(contenttypes) == 1 else "", identitiestype.name, json.dumps(parameters)))
        self.tagstype = tagstype
        self.indextype = indextype
        self.contenttypes = contenttypes
        self.identitiestype = identitiestype
        self.parameters = parameters

    def UnionArrayOf(self):
        if self.tagstype.dtype.bitwidth == 8 and self.tagstype.dtype.signed:
            if self.indextype.dtype.bitwidth == 32 and self.indextype.dtype.signed:
                return awkward1.layout.UnionArray8_32
            elif self.indextype.dtype.bitwidth == 32:
                return awkward1.layout.UnionArray8_U32
            elif self.indextype.dtype.bitwidth == 64 and self.indextype.dtype.signed:
                return awkward1.layout.UnionArray8_64
            else:
                raise AssertionError("no UnionArray* type for index array: {0}".format(self.indextype))
        else:
            raise AssertionError("no UnionArray* type for tags array: {0}".format(self.tagstype))

    def tolayout(self, lookup, pos, fields):
        tags = self.IndexOf(self.tagstype)(lookup.arrays[lookup.positions[pos + self.TAGS]])
        index = self.IndexOf(self.indextype)(lookup.arrays[lookup.positions[pos + self.INDEX]])
        contents = []
        for i, contenttype in enumerate(self.contenttypes):
            layout = contenttype.tolayout(lookup, lookup.positions[pos + self.CONTENTS + i], fields)
            contents.append(layout)
        return self.UnionArrayOf()(tags, index, contents, parameters=self.parameters)

    def hasfield(self, key):
        return any(x.hasfield(key) for x in self.contenttypes)

    def getitem_at(self, viewtype):
        if not all(isinstance(x, RecordArrayType) for x in self.contenttypes):
            raise TypeError("union types cannot be accessed in Numba")

    def getitem_range(self, viewtype):
        if not all(isinstance(x, RecordArrayType) for x in self.contenttypes):
            raise TypeError("union types cannot be accessed in Numba")

    def getitem_field(self, viewtype, key):
        if not all(isinstance(x, RecordArrayType) for x in self.contenttypes):
            raise TypeError("union types cannot be accessed in Numba")

    def lower_getitem_at(self, context, builder, rettype, viewtype, viewval, viewproxy, attype, atval, wrapneg, checkbounds):
        raise NotImplementedError(type(self).__name__ + ".lower_getitem_at not implemented")

    def lower_getitem_range(self, context, builder, rettype, viewtype, viewval, viewproxy, start, stop, wrapneg):
        raise NotImplementedError(type(self).__name__ + ".lower_getitem_range not implemented")

    def lower_getitem_field(self, context, builder, viewtype, viewval, viewproxy, key):
        raise NotImplementedError(type(self).__name__ + ".lower_getitem_field not implemented")
